import math
import random
import numpy as np
from sklearn.neighbors import NearestNeighbors

from exlab.interface.graph import Graph
from exlab.interface.serializer import Serializable
from exlab.utils.io import parameter

from . import operations


class ContextSpatialization(Serializable):
    MAX_AREAS = 100
    NN_NUMBER = 10
    CENTER_RADIUS = 0.27
    MINIMUM_POINTS = 10
    THRESHOLD_ADD = 0.01
    THRESHOLD_DEL = 0.01
    THRESHOLD_ADD_POINT = 0.002
    THRESHOLD_DEL_POINT = 0.002
    THRESHOLD_RESET = 0.05
    THRESHOLD_VALID = 0.5
    THRESHOLD_CANT_CREATE = 0.1
    MIN_DISTANCE_CENTERS = 0.05

    def __init__(self, model, space, boolean=False):
        self.model = model
        self.space = space
        self.evaluatedSpace = self.model.contextSpace
        self.dim = self.evaluatedSpace.dim
        self.boolean = boolean

        # if 

        self.stability = 0
    
        self.resetAreas(False)
    
    def _serialize(self, serializer):
        dict_ = serializer.serialize(
            self, ['number', 'data', 'relevances'])
        return dict_

    @classmethod
    def _deserialize(cls, dict_, serializer, obj=None):
        if obj is None:
            raise Exception('No full deserializer is available for this class')
        return super()._deserialize(dict_, serializer, obj)

    def _postDeserialize(self, dict_, serializer):
        super()._postDeserialize(dict_, serializer)
        serializer = serializer.clone(values={'.area.manager': self})

        self.number = serializer.deserialize(dict_.get('number', []))
        self.data = serializer.deserialize(dict_.get('data', []))
        self.relevances = serializer.deserialize(dict_.get('relevances', []))

    @property
    def dataset(self):
        return self.model.dataset
    
    # def continueFrom(self, cs):
    #     for area in cs.areas:
    #         self._addArea(area.cloneTo(self))

    def _addArea(self, column, point, relevant):
        if self.number[column] >= self.MAX_AREAS:
            return
            # self.data[column, :-1, :] = self.data[column, 1:, :]
            # self.relevances[column, :-1] = self.relevances[column, 1:]

        self._updateArea(column, point, relevant, self.number[column])
        self.number[column] += 1

        # self.findAllPointAreas()
        self.model.invalidateCompetences()
    
    def _updateArea(self, column, point, relevant, index):
        if index == 0:
            return

        self.data[column, index, :] = point.npPlain()
        self.relevances[column, index] = relevant

        self.model.invalidateCompetences()

    # def _removeArea(self, area):
    #     column = area.column
    #     index = self.areas[column].index(area)
    #     self.centers[column, index:-1, :] = self.centers[column, index+1:, :]
    #     self.areas[column].remove(area)

    #     # self.findAllPointAreas()
    #     self.model.invalidateCompetences()

    def resetAreas(self, defaultRelevance=False):
        self.number = np.ones(self.dim, dtype=np.int)
        self.data = np.zeros((self.dim, self.MAX_AREAS, self.space.dim))
        self.relevances = np.full((self.dim, self.MAX_AREAS), defaultRelevance, dtype=np.bool)

        self.model.invalidateCompetences()
    
    def allTrue(self):
        self.resetAreas(True)

    def allFalse(self):
        self.resetAreas(False)
    
    def _groundTruthLidar(self, point, column):
        if point.norm() < 0.2:
            return True
        point = point.plain()
        direction = np.arctan2(point[1], point[0]) / np.math.pi * 180.
        col = int(np.round(direction / 45)) % 8
        # colUp = int(np.ceil(direction / 45)) % 8
        # colDown = int(np.floor(direction / 45)) % 8
        if abs(col - column) <= 2 or abs(col - column) >= 6:
            return True
        # if column in [colUp, colDown]:
        #     return 1, math.inf
        else:
            return False
    
    def _findAreaForOneColumn(self, point, column, space, projection=False, hardcoded=False):

        if projection:
            point = point.projection(parameter(space, self.space))
        # HARDCODED
        # return 1, math.inf
        if hardcoded:
            return int(self._groundTruthLidar(point, column)), math.inf
        # END HARDCODED

        if self.number[column] == 1:
            return 0, math.inf

        if projection:
            point = point.projection(parameter(space, self.space))
        nearest, d = operations.nearestNeighborsFromDataContiguous(self.data[column, 1:self.number[column]], point.npPlain())
        if d[0] > self.space.maxDistance * self.CENTER_RADIUS:
            return 0, math.inf
        return nearest[0] + 1, d[0]

    def findAreaForOneColumn(self, point, column, space=None):
        return self._findAreaForOneColumn(point, column, space, True)[0]

    def findAreaForAllColumns(self, point, space=None):
        point = point.projection(parameter(space, self.space))
        return [self._findAreaForOneColumn(point, column, space)[0] for column in range(self.dim)]
    
    def findAreaForAllColumnsDistances(self, point, space=None):
        point = point.projection(parameter(space, self.space))
        r = [self._findAreaForOneColumn(point, column, space) for column in range(self.dim)]
        return [item[0] for item in r], [item[1] for item in r]
    
    def columns(self, point, space=None):
        point = point.projection(parameter(space, self.space))
        relevances = np.array([False, True])
        return np.array([relevances[self._findAreaForOneColumn(point, column, space, hardcoded=True)[0]] for column in range(self.dim)])
        return np.array([self.relevances[column, self._findAreaForOneColumn(point, column, space)[0]] for column in range(self.dim)])
    
    def columnsAreas(self, areas):
        return np.array([self.relevances[column, area] for column, area in enumerate(areas)])
    
    def findAllAreasForColumn(self, column):
        if self.number[column] == 1:
            return [0] * self.space.getData().shape[0]

        nbrs = NearestNeighbors(n_neighbors=1, algorithm='ball_tree').fit(self.data[column, 1:self.number[column]])
        data = self.space.getData()
        dists, indices = nbrs.kneighbors(data)
        dists, indices = dists.flatten(), indices.flatten() + 1
        indices[dists > self.space.maxDistance * self.CENTER_RADIUS] = 0
        return indices
    
    def update(self, point):
        self.addPoint(point)
        self.addPoint(self.space.randomPoint())

    def addPoint(self, point, test=False):
        # if not test:
        #     return
        if self.space.number < self.MINIMUM_POINTS:
            return

        noiseMargin = 0.002

        probality = max(0.2, np.exp(-self.stability * 0.05))
        if random.uniform(0, 1) < probality:
            self.stability += 1

            point = point.projection(self.space)
            nearest, _ = self.space.nearest(point, n=self.NN_NUMBER)

            areas, distances = self.findAreaForAllColumnsDistances(point)

            currentColumns = self.columnsAreas(areas)

            errors = self.model.forwardErrors(onlyIds=nearest, precise=True, contextColumns=currentColumns)
            c = self.model.computeCompetence(np.mean(errors))
            # c = self.model.competence(onlyIds=nearest, precise=True, contextColumns=currentColumns)

            # print(columns)
            columns = np.arange(self.evaluatedSpace.dim)
            bestAdd = ()
            bestDel = ()
            checkDeletion = False
            if np.any(currentColumns) and random.uniform(0, 1) < 0.5:
                checkDeletion = True

            # print(currentColumns)
            for column in columns:
                if currentColumns[column] == checkDeletion:
                    # print(nearest)
                    # print(self.space.getData(nearest))
                    # print(self.model.outcomeSpace.getData(nearest))
                    # print(currentColumns)
                    # print(self.evaluatedSpace.getData(nearest)[:, currentColumns])
                    # if c is None:

                    # print(c)

                    newColumns = np.copy(currentColumns)
                    newColumns[column] = not newColumns[column]
                    # newColumns[(column + 1) % 8] = not newColumns[(column + 1) % 8]
                    # newColumns[(column - 1) % 8] = not newColumns[(column - 1) % 8]

                    # newColumnsOverwrite = None
                    newColumnsOverwrite = np.full_like(currentColumns, True)
                    newColumnsOverwrite[column] = False
                    # print(newColumns)
                    # print(self.evaluatedSpace.getData(nearest)[:, newColumns])
                    newErrors = self.model.forwardErrors(onlyIds=nearest, contextColumns=newColumns, precise=True, contextColumnsOverwrite=newColumnsOverwrite)
                    nc = self.model.computeCompetence(np.mean(newErrors))
                    # nc = self.model.competence(onlyIds=nearest, contextColumns=newColumns, precise=True, contextColumnsOverwrite=newColumnsOverwrite)

                    progress = -(newErrors - errors)
                    progress = np.sign(progress) * (np.clip(np.abs(progress) - noiseMargin, 0., None))

                    pointProgress = progress[0]
                    meanProgress = np.mean(progress)
                    meanRegression = np.percentile(progress, 10)
                    pc = nc - c
                    progressValue = meanProgress + meanRegression + pc * 0.02

                    # should = self._groundTruthLidar(point, column)
                    if test > 1:
                        print(f'{column}: ={newColumns[column]} ==== {c}+{progressValue} ({meanRegression} {meanRegression} {pc})')
                        for error, nerror, pr, ptx, ctx in zip(errors, newErrors, progress, self.space.getData(nearest), self.evaluatedSpace.getData(nearest)):
                            print(f'{ptx} | {ctx[newColumns]} -> {error:.3f}->{nerror:.3f} {pr:.3f}')
                    # print(progressValue)
                    # addition = newColumns[column]

                    if not checkDeletion and pointProgress >= self.THRESHOLD_ADD_POINT and progressValue >= self.THRESHOLD_ADD and (not bestAdd or progressValue > bestAdd[0]):
                        bestAdd = (progressValue, column, newColumns)
                    elif checkDeletion and pointProgress >= self.THRESHOLD_DEL_POINT and progressValue >= self.THRESHOLD_DEL and (not bestDel or progressValue > bestDel[0]):
                        bestDel = (progressValue, column, newColumns)

                    # registerColumns = None
                    # if (addition and p >= self.THRESHOLD_ADD) or (not addition and p >= self.THRESHOLD_DEL):
                    #     registerColumns = newColumns
                    # # else:
                    # #     registerColumns = currentColumns
                    # if registerColumns is not None:
                    #     canCreateNew = distances[column] >= self.space.maxDistance * self.MIN_DISTANCE_CENTERS
                    #     print(f'Create new columns for {column}: {registerColumns[column]} around {point.plain()} {p}')

                    #     if canCreateNew:
                    #         self._addArea(column, point, registerColumns[column])
                    #     else:
                    #         self._updateArea(column, point, registerColumns[column], areas[column])

                # if (bestAdd or bestDel) and not fullCompAllFalse:
                #     columnsFalse = np.full(self.evaluatedSpace.dim, False)
                #     fullCompAllFalse = self.model.competence(
                #         precise=True, contextColumns=columnsFalse)

                #     columnsTrue = np.full(self.evaluatedSpace.dim, True)
                #     fullCompAllTrue = self.model.competence(
                #         precise=True, contextColumns=columnsTrue)
                    
                #     bestFullComp = max(fullCompAllFalse, fullCompAllTrue)

            for best, deletion, verb in ((bestAdd, False, 'add'),): #, (bestDel, True, 'del')):
                if best:
                    self.stability = 0
                    progressValue, column, newColumns = best
                    # print(f'Should {verb} context column {column} (+{progressValue:.2f}) around {point}')

                    canCreateNew = distances[column] >= self.space.maxDistance * self.MIN_DISTANCE_CENTERS
                    # print(f'Create new columns for {column}: {newColumns[column]} around {point.plain()} +{progressValue} new: {canCreateNew}')

                    # should = self._groundTruthLidar(point, column)
                    # if not should:
                    #     print(f'!! Invalid choice of column {column} for {point}#{nearest[0]} +{progressValue}!\n{self.columns(point)}')
                    #     return 1, 0

                    if canCreateNew:
                        self._addArea(column, point, newColumns[column])
                    else:
                        self._updateArea(column, point, newColumns[column], areas[column])
            
            # return 0, 1 if bestAdd else 0
            # if not bestAdd:
            #     print(f'>> No new column for {point}#{nearest[0]}!\n{self.columns(point)}')


            # for best, deletion, verb in ((bestAdd, False, 'add'), (bestDel, False, 'del')):
            #     if best:
            #         p, column, newColumns = best
            #         print(f'Should {verb} context column {column} ({c:.2f}+{p:.2f}) around {point}')
            #         newArea = None

            #         area = areas[column]
            #         # fullComp = self.model.competence(precise=True)
            #         canCreateNew = distances[column] >= self.space.maxDistance * self.MIN_DISTANCE_CENTERS
            #         if not area.default and area.attempt(currentColumns, newColumns, deletion, canCreateNew):
            #             print('Updating current area')
            #             # previousUsed = area.used
            #             area.used = newColumns[column]
            #         else:
            #             if not canCreateNew:
            #                 print('To close!')
            #                 continue
            #             print('Creating a new area')
            #             newArea = ContextArea(self, point, column, newColumns[column])
            #             self._addArea(newArea)

            #         # newFullComp = self.model.competence(precise=True)
            #         # print(f'Variation: {newFullComp - fullComp}')
            #         # if newFullComp < fullComp * self.THRESHOLD_VALID:  # bestFullComp - self.THRESHOLD_RESET:
            #         #     print('Reverting...')
            #         #     if newArea:
            #         #         self._removeArea(newArea)
            #         #     else:
            #         #         area.used = previousUsed
            #         #     currentColumns = newColumns
            #         # else:
            #         area.stability = 0
    
    # Visual
    def visualizeAreaColumn(self, column=0, options={}):
        g = Graph(title=f'Context areas from {self.space} for column {column}', options=options)

        # All points
        areas = self.findAllAreasForColumn(column)
        points = self.space.getData()

        for relevant in (True, False):
            data = points[np.argwhere(self.relevances[column, areas] == relevant).flatten(), :]
            if len(data) > 0:
                g.scatter(data, label=relevant, color=('orange' if relevant else 'gray'))

        # Centroids
        for (relevant, name, color) in ((True, 'Relevant', 'red'), (False, 'Non relevant', 'black')):
            data = np.argwhere(self.relevances[column, :self.number[column]] == relevant).flatten()
            if not relevant:
                data = data[1:]
            if len(data) > 0:
                g.scatter(self.data[column, data, :], label=f'{name} centers', color=color, marker='x')
        return g

    def visualizeAreaColumns(self, options={}):
        g = Graph(title=f'Context areas from {self.space}', options=options)

        points = self.space.getData()
        offsetWidth = self.model.actionSpace.maxDistance * 0.01
        offsetN = np.math.ceil(np.math.sqrt(self.dim))

        for column in range(self.dim):
            offset = np.array([offsetWidth * (-offsetN / 2 + column % offsetN),
                               offsetWidth * (-offsetN / 2 + column // offsetN)])

            areas = self.findAllAreasForColumn(column)
            data = points[np.argwhere(self.relevances[column, areas]).flatten(), :] + offset
            if len(data) > 0:
                g.scatter(data, label=column, alpha=0.9, marker='.')
        return g


class ContextSpatializationAreas(Serializable):
    MAX_AREAS = 100
    NN_NUMBER = 20
    MINIMUM_POINTS = 100
    THRESHOLD_ADD = 0.05
    THRESHOLD_DEL = 0.
    THRESHOLD_RESET = 0.05
    THRESHOLD_VALID = 0.5
    THRESHOLD_CANT_CREATE = 0.1
    MIN_DISTANCE_CENTERS = 0.05

    def __init__(self, model, space, boolean=False):
        self.model = model
        self.space = space
        self.evaluatedSpace = self.model.contextSpace
        self.dim = self.evaluatedSpace.dim
        self.boolean = boolean

        self.defaultAreas = [ContextArea(self, self.space.zero(), i, False, default=True) for i in range(self.dim)]
        self.resetAreas()

    def _serialize(self, serializer):
        dict_ = serializer.serialize(
            self, ['areas', 'defaultAreas'])
        return dict_

    @classmethod
    def _deserialize(cls, dict_, serializer, obj=None):
        if obj is None:
            raise Exception('No full deserializer is available for this class')
        return super()._deserialize(dict_, serializer, obj)

    def _postDeserialize(self, dict_, serializer):
        super()._postDeserialize(dict_, serializer)
        serializer = serializer.clone(values={'.area.manager': self})

        for areasDict in dict_.get('areas', []):
            for areaDict in areasDict:
                area = serializer.deserialize(areaDict)
                self._addArea(area)
        self.defaultAreas = serializer.deserialize(dict_.get('defaultAreas', []))

    @property
    def dataset(self):
        return self.model.dataset
    
    # def continueFrom(self, cs):
    #     for area in cs.areas:
    #         self._addArea(area.cloneTo(self))

    def _addArea(self, area):
        column = area.column
        if len(self.areas[column]) >= self.MAX_AREAS:
            return

        self.centers[column, len(self.areas[column]), :] = area.center.npPlain()
        self.areas[column].append(area)

        self.findAllPointAreas()
        self.model.invalidateCompetences()

    def _removeArea(self, area):
        column = area.column
        index = self.areas[column].index(area)
        self.centers[column, index:-1, :] = self.centers[column, index+1:, :]
        self.areas[column].remove(area)

        self.findAllPointAreas()
        self.model.invalidateCompetences()

    def resetAreas(self):
        self.areas = [[] for _ in range(self.dim)]
        self.centers = np.zeros((self.dim, self.MAX_AREAS, self.space.dim))

        self.model.invalidateCompetences()

    def allTrue(self, value=True):
        self.resetAreas()
        for column in range(self.dim):
            self._addArea(ContextArea(self, self.space.zero(), column, True))
    
    def allFalse(self):
        self.resetAreas()
    
    def _findAreaForOneColumn(self, point, column, space, projection=False):
        if not self.areas[column]:
            return self.defaultAreas[column], math.inf

        if projection:
            point = point.projection(parameter(space, self.space))
        nearest, d = operations.nearestNeighborsFromDataContiguous(self.centers[column, :len(self.areas[column])], point.npPlain())
        return self.areas[column][nearest[0]], d[0]

    def findAreaForOneColumn(self, point, column, space=None):
        return self._findAreaForOneColumn(point, column, space, True)[0]

    def findAreaForAllColumns(self, point, space=None):
        point = point.projection(parameter(space, self.space))
        return [self._findAreaForOneColumn(point, column, space)[0] for column in range(self.dim)]
    
    def findAreaForAllColumnsDistances(self, point, space=None):
        point = point.projection(parameter(space, self.space))
        r = [self._findAreaForOneColumn(point, column, space) for column in range(self.dim)]
        return [item[0] for item in r], [item[1] for item in r]
    
    def columns(self, goal, space=None):
        areas = self.findAreaForAllColumns(goal, space)
        return np.array([area.used for area in areas])

    def columnsAreas(self, areas):
        return np.array([area.used for area in areas])
    
    def findAllPointAreas(self):
        for column in range(self.dim):
            if self.areas[column]:
                nbrs = NearestNeighbors(n_neighbors=1, algorithm='ball_tree').fit(self.centers[column, :len(self.areas[column])])
                data = self.space.getData()
                _, indices = nbrs.kneighbors(data)
                indices = indices.flatten()
                for i, area in enumerate(self.areas[column]):
                    area.ids = self.space.getIds()[np.argwhere(indices == i).flatten()]
    
    def update(self, point):
        self.addPoint(point)
        self.addPoint(self.space.randomPoint())

    def addPoint(self, point):
        # return
        if self.space.number < self.MINIMUM_POINTS:
            return

        point = point.projection(self.space)
        nearest, _ = self.space.nearest(point, n=self.NN_NUMBER)
        id_ = nearest[0]
        # self.model.saveRestrictionIds(nearest)

        # print('---')
        # print(c)
        # print(point)
        bestAdd = ()
        bestDel = ()
        # print(point)

        areas, distances = self.findAreaForAllColumnsDistances(point)

        # dims = np.arange(self.evaluatedSpace.dim)
        # np.random.shuffle(areas)
        # n = random.randint(1, min(1 + int(self.evaluatedSpace.dim * 0.4), 3))
        # selectedAreas = areas[:n]
        selectedAreas = areas

        # print(areas)
        for area in selectedAreas:
            area.addPoint(id_)
            area.stability += 1
            column = area.column

            probality = 2.  # max(0.05, np.exp(-stability * 0.1))
            # print(self.model, stability, probality)

            # fullCompAllFalse = None
            # fullCompAllTrue = None
            # if area and random.uniform(0, 1) < probality * 0.1:
            #     fullComp = self.model.competence(precise=True)

            #     columnsFalse = np.full(self.evaluatedSpace.dim, False)
            #     fullCompAllFalse = self.model.competence(precise=True, contextColumns=columnsFalse)

            #     columnsTrue = np.full(self.evaluatedSpace.dim, True)
            #     fullCompAllTrue = self.model.competence(precise=True, contextColumns=columnsTrue)

            #     bestFullComp = max(fullCompAllFalse, fullCompAllTrue)
            #     columns = columnsTrue if fullCompAllTrue > fullCompAllFalse else columnsFalse

            #     if fullComp < bestFullComp - self.THRESHOLD_RESET * 2:
            #         # print('Reset all areas!')
            #         self.resetAreas()
            #         area = ContextArea(self, point, columns)
            #         self._addArea(area)

            c = None
            if random.uniform(0, 1) < probality:
                # print(nearest)
                # print(self.space.getData(nearest))
                # print(self.model.outcomeSpace.getData(nearest))
                # print(currentColumns)
                # print(self.evaluatedSpace.getData(nearest)[:, currentColumns])
                if c is None:
                    currentColumns = self.columnsAreas(areas)
                    c = self.model.competence(onlyIds=nearest, precise=True)
                # print(c)

                columns = np.copy(currentColumns)
                columns[column] = not columns[column]
                # print(columns)
                # print(self.evaluatedSpace.getData(nearest)[:, columns])
                nc = self.model.competence(onlyIds=nearest, contextColumns=columns, precise=True)
                p = nc - c
                # print(p)
                if columns[column] and p >= self.THRESHOLD_ADD and (not bestAdd or p > bestAdd[0]):
                    bestAdd = (p, column, columns)
                if not columns[column] and p >= self.THRESHOLD_DEL and (not bestDel or p > bestDel[0]):
                    bestDel = (p, column, columns)

            # if (bestAdd or bestDel) and not fullCompAllFalse:
            #     columnsFalse = np.full(self.evaluatedSpace.dim, False)
            #     fullCompAllFalse = self.model.competence(
            #         precise=True, contextColumns=columnsFalse)

            #     columnsTrue = np.full(self.evaluatedSpace.dim, True)
            #     fullCompAllTrue = self.model.competence(
            #         precise=True, contextColumns=columnsTrue)
                
            #     bestFullComp = max(fullCompAllFalse, fullCompAllTrue)

        for best, deletion, verb in ((bestAdd, False, 'add'), (bestDel, False, 'del')):
            if best:
                p, column, newColumns = best
                print(f'Should {verb} context column {column} ({c:.2f}+{p:.2f}) around {point}')
                newArea = None

                area = areas[column]
                # fullComp = self.model.competence(precise=True)
                canCreateNew = distances[column] >= self.space.maxDistance * self.MIN_DISTANCE_CENTERS
                if not area.default and area.attempt(currentColumns, newColumns, deletion, canCreateNew):
                    print('Updating current area')
                    # previousUsed = area.used
                    area.used = newColumns[column]
                else:
                    if not canCreateNew:
                        print('To close!')
                        continue
                    print('Creating a new area')
                    newArea = ContextArea(self, point, column, newColumns[column])
                    self._addArea(newArea)

                # newFullComp = self.model.competence(precise=True)
                # print(f'Variation: {newFullComp - fullComp}')
                # if newFullComp < fullComp * self.THRESHOLD_VALID:  # bestFullComp - self.THRESHOLD_RESET:
                #     print('Reverting...')
                #     if newArea:
                #         self._removeArea(newArea)
                #     else:
                #         area.used = previousUsed
                #     currentColumns = newColumns
                # else:
                area.stability = 0
    
    # Visual
    def visualizeAreaColumn(self, column=0, options={}):
        g = Graph(title=f'Context areas from {self.space} for column {column}', options=options)
        areas = {}
        for area in self.areas[column]:
            if area.used in areas.keys():
                areas[area.used] = np.vstack((areas[area.used], self.space.getData(area.ids)))
            else:
                areas[area.used] = self.space.getData(area.ids)
        for used, data in areas.items():
            g.scatter(data, label=used, color=('orange' if used else 'gray'))
        g.scatter(self.centers[column, :len(self.areas[column]), :], label='Center', color='red', marker='x')
        return g

    def visualizeAreaColumns(self, options={}):
        g = Graph(title=f'Context areas from {self.space}', options=options)
        areas = {}
        allAreas = [b for a in self.areas for b in a]

        offsetWidth = self.model.actionSpace.maxDistance * 0.01
        offsetN = np.math.ceil(np.math.sqrt(self.dim))
        for area in allAreas:
            if area.used:
                offset = np.array([offsetWidth * (-offsetN / 2 + area.column % offsetN),
                                   offsetWidth * (-offsetN / 2 + area.column // offsetN)])
                if area.column in areas.keys():
                    areas[area.column] = np.vstack((areas[area.column], self.space.getData(area.ids) + offset))
                else:
                    areas[area.column] = self.space.getData(area.ids) + offset
        for column, data in areas.items():
            g.scatter(data, label=column, alpha=0.9, marker='.')
        return g


class ContextArea(Serializable):
    def __init__(self, manager, center, column, used, default=False):
        self.manager = manager
        self.center = center
        self.column = column
        self.used = used
        self.ids = np.array([])
        self.stability = 0
        self.default = default
    
    def _serialize(self, serializer):
        dict_ = serializer.serialize(
            self, ['center', 'column', 'stability', 'default'])
        return dict_

    @classmethod
    def _deserialize(cls, dict_, serializer, obj=None):
        if obj is None:
            obj = cls(serializer.get('.area.manager'),
                      serializer.deserialize(dict_.get('center')),
                      serializer.deserialize(dict_.get('column')),
                      serializer.deserialize(dict_.get('used')),
                      serializer.deserialize(dict_.get('default'))
                      )
        return super()._deserialize(dict_, serializer, obj)

    def _postDeserialize(self, dict_, serializer):
        super()._postDeserialize(dict_, serializer)
        self.stability = dict_.get('stability', 0)
    
    # def cloneTo(self, manager):
    #     columns = np.full(manager.evaluatedSpace.dim, False)

    #     posn = 0
    #     poso = 0
    #     for sn in manager.evaluatedSpace.cols:
    #         for so in self.manager.evaluatedSpace.cols:
    #             if sn.matches(so):
    #                 columns[posn:posn + sn.dim] = self.columns[poso:poso + so.dim]
    #                 break
    #             poso += so.dim
    #         posn += sn.dim

    #     new = self.__class__(manager, self.center, columns)
    #     return new

    def addPoint(self, id_):
        if not self.default:
            self.ids = np.append(self.ids, id_)
    
    def attempt(self, columns, newColumns, deletion, canCreateNew, precise=True):
        c = self.manager.model.competence(
            onlyIds=self.ids, contextColumns=columns, precise=precise)
        nc = self.manager.model.competence(
            onlyIds=self.ids, contextColumns=newColumns, precise=precise)
        print(f'Attempt on current region: p={nc - c}')
        return nc - c > (ContextSpatialization.THRESHOLD_DEL if deletion else ContextSpatialization.THRESHOLD_ADD) - \
                (0. if canCreateNew else ContextSpatialization.THRESHOLD_CANT_CREATE)


    def __repr__(self):
        if self not in self.manager.areas:
            return f'Area centered on {self.center}'
        return f'Area {self.manager.areas.index(self)} centered on {self.center}'



from .live_entity import LiveEntity


class PhysicalEntity(LiveEntity):
    PHYSICAL = True

    def __init__(self, kind, name='', manager=None):
        super().__init__(kind, name, manager=manager)
        self.shape = None

    def _activate(self):
        super()._activate()
        self.initPhysics(self.engine.physics)

    def _deactivate(self):
        super()._deactivate()
        self.stopPhysics(self.engine.physics)

    def initPhysics(self, physics):
        return None

    def stopPhysics(self, physics):
        pass



from ..engine import Engine, GraphicalEngine, DrawOptions


class InternalEngine(Engine):
    PHYSICAL = True

    # Physics
    def updatePhysics(self, dt=0.00000001):
        self._updatePhysics(dt)
        self.iteration += 1
        self.time += dt



from panda3d.core import Point3, Vec3, Vec4, Point2, LPoint2i, Geom, GeomNode, GeomTriangles, GeomVertexFormat, GeomVertexData, GeomVertexWriter

import numpy as np


class PandaElement(object):
    def __init__(self, pt):
        self.pt = pt
        self.pt.elements.append(self)

        self.written = False
        self.color = Vec4(self.pt.color)
        self.normal = None
        self.normalFirst2 = None

    def alpha(self, alpha):
        self.color.setZ(alpha)
    
    def findNormal(self, a, b, c):
        ab = b - a
        ac = c - a
        # print('===')
        # print(ab)
        # print(ac)
        return ab.cross(ac).normalized()
    
    def writeData(self, points, normals, addPrimitives=True):
        self.written = True
        if not isinstance(normals, (list, tuple)):
            normals = [normals] * len(points)

        index = self.pt.vertex.getWriteRow()
        for point, normal in zip(points, normals):
            self.pt.vertex.addData3f(point)
            self.pt.vnormal.addData3f(normal)
            self.pt.vcolor.addData4f(self.color)
        
        if addPrimitives:
            self.pt.primitive.addVertices(index, index + 1, index + 2)
            if len(points) >= 4:
                self.pt.primitive.addVertices(index, index + 2, index + 3)

        return index

class PandaElementTriangle(PandaElement):
    def setup(self, a, b, c, reverse=False):
        if not reverse:
            self.a, self.b, self.c = a, b, c
        else:
            self.a, self.b, self.c = c, b, a
        return self
    
    def write(self):
        points = [self.a, self.b, self.c]
        self.normal = self.findNormal(self.a, self.b, self.c)

        self.writeData(points, self.normal)

class PandaElementPlane(PandaElement):
    def setup(self, a, b, c, d, reverse=False):
        if not reverse:
            self.a, self.b, self.c, self.d = a, b, c, d
        else:
            self.a, self.b, self.c, self.d = d, c, b, a
        return self
    
    def write(self):
        points = [self.a, self.b, self.c, self.d]

        self.normal = self.findNormal(self.a, self.b, self.c)
        # print(self.normal)
        normals = [self.normal] * 4
        if self.normalFirst2:
            normals[0], normals[1] = [self.normalFirst2, self.normalFirst2]

        self.writeData(points, normals)


class PandaTools(object):
    def __init__(self, autoStartPrimitive=True):
        if autoStartPrimitive:
            self.startPrimitive()
        self.elements = []
        self.color = Vec4(1., 1., 1., 1.)

    def plane(self, a, b, c, d, **options):
        return [PandaElementPlane(self).setup(a, b, c, d)]
    
    def horizontalPlane(self, a, c, **options):
        b = Point3(c.x, a.y, a.z)
        d = Point3(a.x, c.y, c.z)
        return [PandaElementPlane(self).setup(a, b, c, d)]
    
    def verticalPlane(self, a, c, **options):
        b = Point3(c.x, c.y, a.z)
        d = Point3(a.x, a.y, c.z)
        return [PandaElementPlane(self).setup(a, b, c, d)]
    
    def cube(self, a, c, z, **options):
        b = Point3(c.x, a.y, a.z)
        d = Point3(a.x, c.y, c.z)
        return self.quadri(a, b, c, d, z, **options)

    def quadri(self, a, b, c, d, z, **options):
        def z0(p):
            return Point3(p.x, p.y, z)

        return [PandaElementPlane(self).setup(a, b, c, d),
                PandaElementPlane(self).setup(z0(a), z0(b), z0(c), z0(d), reverse=True),
                self.verticalPlane(c, z0(b))[0],
                self.verticalPlane(d, z0(c))[0],
                self.verticalPlane(a, z0(d))[0],
                self.verticalPlane(b, z0(a))[0]]
    
    def cylinder(self, radius, height, origin=Point3.zero(), steps=50):
        elements = []
        upCenter = origin + Vec3(0, 0, height)
        for i in range(steps):
            a = origin + Vec3(np.cos(np.math.pi * 2 / steps * i) * radius,
                              np.sin(np.math.pi * 2 / steps * i) * radius, 0)
            c = origin + Vec3(np.cos(np.math.pi * 2 / steps * (i + 1)) * radius,
                              np.sin(np.math.pi * 2 / steps * (i + 1)) * radius, height)
            elements.append(self.verticalPlane(a, c)[0])
            elements.append(PandaElementTriangle(self).setup(upCenter, Point3(a.x, a.y, a.z + height), c))
            elements.append(PandaElementTriangle(self).setup(origin, Point3(c.x, c.y, c.z - height), a))
        return elements

    def startPrimitive(self):
        self.vdata = GeomVertexData('chunk-vertex', GeomVertexFormat.getV3n3c4(), Geom.UHStatic)

        self.vertex = GeomVertexWriter(self.vdata, 'vertex')
        self.vnormal = GeomVertexWriter(self.vdata, 'normal')
        self.vcolor = GeomVertexWriter(self.vdata, 'color')

        self.primitive = GeomTriangles(Geom.UHStatic)
    
    def endPrimitive(self):
        self.primitive.closePrimitive()
    
    def end(self):
        for element in self.elements:
            if not element.written:
                element.write()
        self.endPrimitive()
        return self.createNode()
    
    def createNode(self):
        self.node = GeomNode('chunk')

        geom = Geom(self.vdata)
        geom.addPrimitive(self.primitive)
        self.node.addGeom(geom)
    
        return self.node



from exlab.interface.serializer import Serializable
import random


class StrategySet(set, Serializable):
    def __init__(self, strategies=[], agent=None):
        super().__init__()
        self.agent = agent
        for strategy in strategies:
            self.add(strategy)

    def _serialize(self, serializer):
        return [serializer.serialize(item) for item in self]

    def add(self, strategy):
        if strategy in self:
            return
        if self.agent:
            assert strategy.agent == self.agent
            # self.agent.addChildModule(strategy)
        set.add(self, strategy)

    def sample(self):
        if not self:
            raise Exception('No strategy available! Please add one first')
        return random.choice(list(self))



import numpy as np
import math
import copy
import multiprocessing as mp

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

from scipy.spatial import cKDTree
from scipy.spatial.distance import euclidean

from exlab.interface.serializer import Serializable
from exlab.interface.graph import Graph
from exlab.modular.module import Module

from dinos.data.data import Goal, SingleData
from dinos.data.path import ActionNotFound

from dinos.utils.move import MoveConfig
# from dinos.utils.maths import iterrange
# from ..utils.io import getVisual, plotData, visualize

from .tests import Test, TestResult


class Evaluation(Serializable):
    def __init__(self, iteration, results, models=[]):
        self.iteration = iteration
        self.results = results
        self.meanError = np.mean([result.meanError for result in self.results])
        self.meanStd = np.mean([result.std for result in self.results])
        self.models = models

    def _serialize(self, serializer):
        dict_ = serializer.serialize(self, ['iteration', 'meanError', 'meanStd', 'results', 'models'])
        return dict_
    
    @classmethod
    def _deserialize(cls, dict_, serializer, obj=None):
        if obj is None:
            obj = cls(dict_.get('iteration'),
                      serializer.deserialize(dict_.get('results', [])))
        return super()._deserialize(dict_, serializer, obj)

    # @classmethod
    # def _deserialize(cls, dict_, options={}, obj=None):
    #     obj = obj if obj else cls(dict_['iteration'],
    #                               dict_['meanError'],
    #                               dict_['meanStd'],
    #                               Serializer.deserialize(
    #                                   dict_['results'], options=options),
    #                               Serializer.deserialize(dict_['models'], options=options))

    #     return obj

    def __repr__(self):
        return f'Evaluation @t={self.iteration} µ={self.meanError:.3f} σ={self.meanStd:.3f} ({len(self.results)} test(s))'


class Evaluator(Module):
    """Evaluation of a learning agent on a testbench."""
    DEFAULT_ERROR = 1.0

    def __init__(self, agent, environment, options={}, method=None):
        """
        agent LearningAgent
        """
        super().__init__('Evaluator', environment, loggerTag='evaluation')

        self.agent = agent
        self.environment = environment
        self.options = options
        self.method = method if method else options.get('method', 'execution')
        self.ts = options.get('ts', [])
        # Will contain all evaluation performed
        # Each evaluation structured as below:
        #    [t, [[[perf_point0, perf_point1,...], sumError_task_0, std_task_0], ...], errors, stds]
        self.evaluations = {}

    def _serialize(self, serializer):
        dict_ = serializer.serialize(
            self, ['options', 'method', 'evaluations'])
        return dict_
    
    @classmethod
    def _deserialize(cls, dict_, serializer, obj=None):
        if obj is None:
            obj = cls(serializer.get('agent'),
                      serializer.get('environment'),
                      options=dict_.get('options', {}),
                      method=dict_.get('method'))

        return super()._deserialize(dict_, serializer, obj)

    def _postDeserialize(self, dict_, serializer):
        super()._postDeserialize(dict_, serializer)
        self.evaluations = serializer.deserialize(dict_.get('evaluations'), default={})
        # for evaluation in dict_.get('evaluations', []):
        #     obj = serializer.deserialize(evaluation)
        #     self.evaluations[obj.iteration] = obj

    # @classmethod
    # def _deserialize(cls, dict_, agent, environment, options={}, obj=None):
    #     obj = obj if obj else cls(
    #         agent, environment, dict_.get('options'), dict_.get('method'))

    #     # operations
    #     obj.evaluations = Serializer.deserialize(
    #         dict_.get('evaluations'), options=options)

    #     return obj

    # def nextEvaluationIteration(self, currentIteration):
    #     return next(t for t in self.ts if t > currentIteration)

    # def evaluateAllTime(self):
    #     """Evaluate the learning agent at different times."""
    #     for t in self.ts:
    #         self.evaluateAllTests(t)

    def evaluate(self):
        """Evaluate performance and record it with time given."""

        # Will contain all evaluation information
        results = []

        self.logger.debug(f'Evaluating all tests...')
        for test in self.environment.tests:
            result = self.evaluateTest(test)
            if result:
                results.append(result)

        if len(results) == 0:
            return None
        
        models = []

        # if self.agent.dataset:
        #     models = [{'competence': model.competence(), 'model': model}
        #               for model in self.agent.dataset.models]
        # else:
        #     models = []

        evaluation = Evaluation(self.environment.iteration, results, models)
        self.evaluations[evaluation.iteration] = evaluation
        self.logger.info(f'Global evaluation results: {evaluation}')

        return evaluation

    def evaluateTest(self, test):
        """Compute evaluation for a given task."""
        if test.id is None:
            raise Exception('A test must be assigned to a scene to be evaluated!')
    
        self.logger.debug(f'Evaluating test {test.name} with method {self.method}...')

        #assert len(self.testbench[task]) > 0

        # space = self.agent.dataset.space(spacename)
        # dataset = self.agent.dataset
        # planner = self.agent.planner
        # performer = self.agent.performer
        #model = self.agent.dataset.findModelByOutcomeSpace(space)

        #if not model:
        #    return

        # Find the size of the task space at the time
        '''exeId = self.agent.epmem_length[t]
        max_i = len(self.agent.dataset.outcomeSpaces[task].data)
        for i in range(len(self.agent.dataset.outcomeSpaces[task].data)):
            if self.agent.dataset.outcomeSpaces[task].ids[i] > exeId:
                max_i = i
                break

        # Retrieve dataset at the time
        data = np.array(self.agent.dataset.outcomeSpaces[task].data[0:max_i])

        # Create KD-Tree for evaluation
        tree = None
        if len(data) > 0:
            tree = cKDTree(data)'''

        #from evaluation.visualization import Visualizer
        #Visualizer([model.get_competence_visualizer(dataset=testbench), model.get_competence_visualizer()]).plot()
        #Visualizer([model.get_competence_visualizer(method=1), model.get_competence_visualizer()]).plot()
        #raw_input()

        # Will contain all evaluation information
        results = []
        # Error sum on the testbench
        # Sum of quadratic error on testbench
        # n_ = 0

        if test.preTest:
            test.preTest()
        else:
            self.environment.setupPreTest(test)

        for rawPoint in test.points:
            if test.prePoint:
                test.prePoint(rawPoint)
            else:
                self.environment.setupPreTestPoint(test, rawPoint)

            if self.agent.dataset:
                point = self.agent.dataset.convertData(rawPoint)
            else:
                point = rawPoint

            self.logger.debug2(f'Evaluating point {point}')

            # print("Goal: {} {}".format(point, space))
            # -- Use distance --
            if self.method == 'distance':
                result = self.evaluateDistance(point)
            # -- Use planning (and perform the planned path) --
            elif self.method == 'planning':
                result = self.evaluatePlanning(point)
            elif self.method == 'execution':
                result = self.evaluateExecution(point)
            '''if tree is None:
                dist = self.agent.dataset.outcomeSpaces[task].options['out_dist']
            else:
                dist, _ = tree.query(point, k=1, eps=0.0, p=2)'''
            #p = std / space.options['max_dist']
            #if p > 10.:
            #    print("{} -> {} {}".format(point, p, competence))
            error = result[0]
            if error >= 0.:
                results.append(result)
        
        testResult = TestResult(test, self.agent.iteration, results, method=self.method)
        
        if test.postTest:
            test.postTest(testResult)

        self.logger.info(f'Test evaluation results: {testResult}')

        #print(eva)
        # errors.sort()
        # we delete aberations
        #n = 10  # len(eva) / 10 + 1
        #if eva > n:
        #    eva = eva[0:-n]

        return testResult

    def evaluateDistance(self, point):
        _, dist = point.space.nearestDistance(point)
        if not dist:
            return Evaluator.DEFAULT_ERROR, point, None
        return min(1., dist[0] / point.space.maxDistance), point, None

    def evaluatePlanning(self, point):
        try:
            _, _, distance = self.agent.planner.planDistance(
                point)  # , hierarchical=False)
            # print(dist)
            error = min(1., distance / point.space.maxDistance)
        except ActionNotFound as e:
            print(f"Failed {point}")
            error = min(1., e.minDistanceReached /
                        point.space.maxDistance) if e.minDistanceReached else self.DEFAULT_ERROR

            #p2, dist2 = space.nearestDistance(point)
            # print("{} {} // {}".format(point, dist, paths))
            '''if dist < space.options['max_dist'] / 10:
                #print(space.options['max_dist'] / 10)
                #print("Test")
                n_ += 1
                paths = p#planner.plan(goal)
                #print(planner.hierarchical)
                #print(planner.chaining)
                #print(len(paths.paths[0]))
                competence, _, dist = dataset.get_competence_std_paths(paths)#model.get_competence_std(point)
                competence = 0.5 + competence / 2.
                #print(competence)
            else:
                competence, dist = (0, dist)#space.options['out_dist'])
                competence = 0.5 - min(1., dist / space.options['max_dist']) / 2.'''

            #error = 0.4 + 0.5 * min(1., dist / space.options['max_dist'])
        return error, point, None

    def evaluateExecution(self, point):
        try:
            point.space._validate()
            discrete = False
            if self.environment.world.discretizeActions:
                print('Wesh!')
                point = Goal(point)
                point.setRelative(False)
                print(point)
                addDistance = 100.
            else:
                addDistance = point.space.maxDistance

            discrete = self.environment.world.discretizeStates
            self.environment.world.discretizeStates = False
            goalDistance = point.relativeData(self.environment.state()).norm()
            absoluteGoal = point.absoluteData(self.environment.state())
            self.environment.world.discretizeStates = discrete

            # print('@@@')
            # print(absoluteGoal)
            # print(point)
            # print(self.environment.state())

            self.agent.reach(MoveConfig(
                goal=point, evaluating=True))
            self.environment.runScheduled(evaluating=True)

            self.environment.world.discretizeStates = False
            difference = absoluteGoal.relativeData(self.environment.state())
            # print(self.environment.state())
            self.environment.world.discretizeStates = discrete

            reached = absoluteGoal - difference
            distance = difference.norm()
            try:
                error = min(1., distance / min(goalDistance, addDistance))
            except Exception:
                error = 1.
            
            self.logger.debug2(
                f'Aiming for {absoluteGoal} and reached {reached}: diff {distance:.3f}, error {error:.3f}')

            # print('DISTANCE', reached, difference, distance, error)
            return error, point, reached
            # paths, _, dist = self.agent.planner.planDistance(point)#, hierarchical=False)
            # spaces = list(env.dataset.outcomeSpaces)
            # for s in spaces:
            #     env.scene.property(s.property).save()
            # results = self.agent.performer.perform(paths, self.agent.dataset)
        except ActionNotFound:
            return Evaluator.DEFAULT_ERROR, point, None
    
    # Visual
    def visualizeEvaluations(self):
        g = Graph()
        g.plot(list(self.evaluations.keys()),
               list(map(lambda x: x.meanError, self.evaluations.values())),
               std=list(map(lambda x: x.meanStd, self.evaluations.values())))
        return g

    # Plots
    # def plot(self):
    #     """Plot evaluation on all iterations."""
    #     x = np.array([eval.iteration for eval in self.evaluations])
    #     error = np.array([eval.meanError for eval in self.evaluations])
    #     std = np.array([eval.meanStd for eval in self.evaluations])
    #
    #     ax.plot(x, y) #, marker=options['marker'], linestyle=options['linestyle'], color=options['color']
    #     # plt.yscale('log')
    #     plt.grid(True)

    # def plot_mean_task(self, task, ax, options):
    #     """Plot evaluation for a specific task space."""
    #     x = []
    #     y = []
    #     for i in self.evaluation:
    #         x.append(i[0])
    #         y.append(i[1][task][1])
    #     ax.plot(np.array(x), np.array(
    #         y), marker=options['marker'], linestyle=options['linestyle'], color=options['color'])
    #     plt.yscale('log')
    #     plt.grid(True)

    # #### The following functions are here ro ease the use of Visualizers

    # def get_complete_visualizer(self, prefix=""):
    #     """Create dictionary to visualize evaluation."""
    #     dico = {}
    #     dico['limits'] = {'min': [None, None], 'max': [None, None]}
    #     dico['title'] = prefix + "Evaluations of the algo"
    #     dico['color'] = ['k']
    #     dico['marker'] = ['']
    #     dico['linestyle'] = ['solid']
    #     dico['axes'] = ['Iterations', 'Mean error on testbench']
    #     dico['legend'] = []
    #     dico['plots'] = [lambda fig, ax, options: self.plot(ax, options)]

    #     return dico

    # def get_task_visualizer(self, task, prefix=""):
    #     """Create dictionary to visualize evaluation of a specific task space."""
    #     dico = {}
    #     dico['limits'] = {'min': [None, None], 'max': [None, None]}
    #     dico['title'] = prefix + "Evaluations of the algo on task " + str(task)
    #     dico['color'] = ['k']
    #     dico['marker'] = ['']
    #     dico['linestyle'] = ['solid']
    #     dico['axes'] = ['Iterations', 'Mean error on testbench']
    #     dico['legend'] = []
    #     dico['plots'] = [lambda fig, ax,
    #                      options: self.plot_mean_task(task, ax, options)]

    #     return dico

    # # Visual
    # def visualizeEvaluation(self, byTest=True, prefix=''):
    #     """Return a dictionary used to visualize outcomes reached for the specified outcome space."""

    #     """Plot evaluation on all iterations."""
    #     # x = np.array([eval.iteration for eval in self.evaluations])
    #     # error = np.array([eval.meanError for eval in self.evaluations])
    #     # std = np.array([eval.meanStd for eval in self.evaluations])
    #     #
    #     # ax.plot(x, y) #, marker=options['marker'], linestyle=options['linestyle'], color=options['color']
    #     # # plt.yscale('log')
    #     # plt.grid(True)

    #     if byTest:
    #         data = np.array([[(eval.iteration, 1. - result.meanError) for result in eval.results]
    #                          for eval in self.evaluations]).transpose(1, 0, 2)
    #         stds = np.array([[result.std for result in eval.results]
    #                          for eval in self.evaluations]).transpose()
    #     else:
    #         data = [np.array([(eval.iteration, 1. - eval.meanError)
    #                           for eval in self.evaluations])]
    #         stds = [np.array([eval.meanStd for eval in self.evaluations])]

    #     def visual(d, std):
    #         return lambda fig, ax, options: plotData(d, fig, ax, options, std=std, lines=True)

    #     return getVisual(
    #         [visual(d, std) for d, std in zip(data, stds)],
    #         # marker='o',
    #         title=prefix + "Global competences"
    #     )

    # def visualizeCompetence(self, prefix=""):
    #     """Return a dictionary used to visualize outcomes reached for the specified outcome space."""
    #     data = {}
    #     for v in self.evaluation:
    #         for model, mdict in v['models'].items():
    #             if model not in data:
    #                 data[model] = []
    #             data[model].append((v['iteration'], mdict['competence']))
    #     plots = [lambda fig, ax, options: plotData(
    #         np.array(d), fig, ax, options, lines=True) for _, d in data.items()]
    #     legend = [str(model) for model, _ in data.items()]
    #     return getVisual(
    #         plots,
    #         marker='.',
    #         legend=legend,
    #         title=prefix + "Models competence"
    #     )

    # # Plot
    # def plot(self, byTest=True):
    #     visualize(self.visualizeEvaluation(byTest=byTest))

    # def plotCompetence(self):
    #     visualize(self.visualizeCompetence())

    # # Api
    # def apiget_evaluation(self, range_=(-1, -1)):
    #     return {'data': iterrange(self.evaluation, range_)}



import sys
import copy
import math
import random
import numpy as np
from scipy.spatial import cKDTree
from scipy.spatial.distance import euclidean

# import matplotlib as plt

from .model import Model

# from ..utils.logging import DataEventHistory, DataEventKind
# from ..utils.io import getVisual, plotData, visualize, concat
# from ..utils.serializer import serialize
# from ..utils.maths import multivariateRegression, multivariateRegressionError
from dinos.utils.maths import multivariateRegression, multivariateRegressionError, linearValue
from dinos.data.data import Data, Goal, Observation, Action
# from dinos.data.abstract import *
from dinos.data.space import SpaceKind
from dinos.data.path import ActionNotFound


class RegressionModel(Model):
    NN_CONTEXT = 40
    NN_CONTEXT_RATIO = 0.05
    NN_ALOCALITY = 5
    NN_LOCALITY = 10
    MINIMUM_LENGTH = 0.001
    MAXIMUM_NULL = 5
    OUTLIER_MAX_DISTANCE = 100
    ENFORCE_MAX_NN_DISTANCE = False
    MAX_ERROR_REACHABLE = 0.002

    def __init__(self, dataset, actionSpace, outcomeSpace, contextSpace=[], restrictionIds=None, model=None,
                 register=True, metaData={}):
        super().__init__(dataset, actionSpace, outcomeSpace,
                         contextSpace, restrictionIds, model, register, metaData)
        self.bestLocalityCandidates = [[], []]
        self.lastCloseIds = None

    # def computeCompetence(self, error, distanceGoal=0):
    #     distanceGoal = min(distanceGoal, 1.)
    #     error = min(error, 0.1)
    #     return max(0, min(1., (1. - distanceGoal - error ** 2) / np.exp((error * 20) ** 3)))

    def reachable(self, goal: Goal, context: Observation = None, precision=0.25, precisionOrientation=0.1, inverse=True, contextColumns=None, dontMove=[], adaptContext=False):
        assert(goal is not None)
        try:
            # norm = goal.norm() / goal.space.maxDistance
            # if norm < precision * 2:
            #     precision = max(precision / 2, norm / 1.5)
            # TODO
            if inverse:
                a0, y0, _, _, _, distance, _, error = self.bestLocality(
                    goal, context, contextColumns=contextColumns, dontMove=dontMove, adaptContext=adaptContext)
            else:
                raise Exception('To check')
                # y0, goalDistanceNormalized = self.bestLocality(
                #     goal, context, getClosestOutcome=True, contextColumns=contextColumns)
                # a0 = None

            if goal.norm() < 0.001:
                a0 = a0.space.zero().setRelative(True)
                y0 = y0.space.zero().setRelative(True)
                distance = 0
                reachable = True
            if error > self.MAX_ERROR_REACHABLE:
                reachable = False

                # self.dataset.logger.info(
                #     f'{reachable}: {goal} and got {y0} ({a0}->) due to error {error} > {self.MAX_ERROR_REACHABLE}')
            else:
                if y0.norm() < 0.001:
                    orientation = 1.
                else:
                    orientation = 1. - goal.npPlain().dot(y0.npPlain()) / (goal.norm() * y0.norm())

                distance = (goal - y0).norm()
                distanceGoalNormalized = max(0, distance - goal.space.maxDistance * 0.01) / goal.norm()
                # reachable = (distanceGoalNormalized < precision and distanceGoalNormalized < 0.7) or distanceGoalNormalized < precision * 0.5
                # self.dataset.logger.info(f'HEY=== {reachable}: {goal} and got {y0} ({a0}->) distance {orientation} {distanceGoalNormalized}')

                ratio = self.outcomeSpace.number / 500
                multiplier = linearValue(1.5, 1, ratio)
                reachable = orientation < precisionOrientation and distanceGoalNormalized < precision * multiplier

                # if not reachable:
                #     self.dataset.logger.info(
                #         f'{reachable}: {goal} and got {y0} ({a0}->) distance {orientation} {distanceGoalNormalized}')
            
            # if error > self.MAX_ERROR_REACHABLE:
            #     reachable = False

            #     self.dataset.logger.info(
            #         f'{reachable}: {goal} and got {y0} ({a0}->) due to error {error} > {self.MAX_ERROR_REACHABLE}')

            return reachable, a0, y0, distance
        except ActionNotFound:
            return False, None, None, None

    def inverse(self, goal: Goal, context: Observation = None, adaptContext=False, contextColumns=None, entity=None):
        assert(goal is not None)
        a0, y0, c0, comp, error, distance = self.bestLocality(
            goal, context, adaptContext=adaptContext, contextColumns=contextColumns, entity=entity)[:6]
        return a0, y0, c0, comp, error, distance  # , dist, distNormalized

    def npForward(self, action: Action, context: Observation = None, bestContext=False, contextColumns=None,
                  ignoreFirst=False, entity=None, debug=False):
        assert(action is not None)

        results = self._nearestData(
            action, context, self.NN_LOCALITY, bestContext, outcome=False, contextColumns=contextColumns,
            nearestUseContext=True, ignoreFirst=ignoreFirst, entity=entity)
        ids, _, context, _, space, action, _, _, actionContextPlain, contextColumns = results

        # print(f'{self.actionContextSpace.getNpPlainPoint(ids)} > {self.outcomeSpace.getNpPlainPoint(ids)}')

        # if debug:
        #     print('=========')
        #     print(action)
        #     print(np.array(self.actionSpace.getPlainPoint(ids)))
        #     print(dist)
        #     print(context)
        # print(self.contextSpace.getPlainPoint(restrictionIds))
        # print("Selecting 2 {}".format(ids))
        # print("Distances 3 {}".format(dists))

        # ids = self.actionSpace.getActionIndex(ids)
        if len(ids) == 0:
            return None, None

        x = space.getNpPlainPoint(ids)
        y = self.outcomeSpace.getNpPlainPoint(ids)
        fullColumns = self.multiContextColumns(contextColumns, space, context)
        weights = space.findWeights(fullColumns, self.contextSpace)
        # c = self.contextSpace.getNpPlainPoint(ids)
        # print(f'{x} {y} {c}')

        # print(contextColumns)
        RegressionModel._db_data = (x, y, actionContextPlain, fullColumns, weights)
        y0, error = multivariateRegressionError(x, y, actionContextPlain, columns=fullColumns, weights=weights)
        maxOutliers = self.outcomeSpace.maxDistance * self.OUTLIER_MAX_DISTANCE
        if np.any(y0 > maxOutliers):
            # print('======EROEOREOROEORROEO======')
            # print(f'!!! {y0} {self.outcomeSpace.maxDistance}')
            y0 = y[-1]
            error = min(error, 0.2)

        return y0, error

    # def adaptContext(self, goal, context=None, relative=True, contextColumns=None, entity=None):
    #     if not context or not self.hasContext(self.contextSpace, contextColumns):
    #         return None

    #     self.outcomeSpace._validate()
    #     self.actionSpace._validate()

    #     goalPlain = Data.npPlainData(goal, self.outcomeSpace)

    #     # Compute best locality candidates
    #     ids, dist = self.outcomeSpace.nearest(goalPlain,
    #                                           n=self.NN_LOCALITY,
    #                                           restrictionIds=self.restrictionIds,
    #                                           otherSpace=self.actionContextSpace)

    #     # # Check if distance to goal is not too important
    #     # minPointsY = 5
    #     # idsPart = np.squeeze(np.argwhere(dist < self.outcomeSpace.maxNNDistance), axis=1)
    #     # if len(idsPart) < minPointsY:
    #     #     idsPart = range(0, len(idsPart))
    #     # ids = ids[idsPart]
    #     #
    #     # if len(ids) == 0:
    #     #     raise ActionNotFound("Not enough points to compute")

    #     print('=================================')
    #     print(dist)
    #     print(ids)
    #     try:
    #         scores = [self.inverse(goal, self.contextSpace.getPoint(id_)[0], contextColumns=contextColumns)[
    #             4] + d / 2 for id_, d in zip(ids, dist)]
    #     except ActionNotFound:
    #         scores = []
    #         for id_, d in zip(ids, dist):
    #             try:
    #                 scores.append(self.inverse(
    #                     goal, self.contextSpace.getPoint(id_)[0], contextColumns=contextColumns)[4] + d / 2)
    #             except ActionNotFound:
    #                 scores.append(-1000)
    #     ids = ids[np.argsort(scores)]
    #     print(scores)
    #     print(ids)

    #     # index
    #     index = ids[0]
    #     c0controllable = self.contextSpace.getPoint(index)[0]

    #     print(c0controllable)
    #     print(self.outcomeSpace.getPoint(index)[0])

    #     # merge
    #     nonControllable = context.projection(self.nonControllableContext(), entity=entity)
    #     c0 = c0controllable.extends(nonControllable)

    #     if relative:
    #         c0 = c0 - context
    #         c0 = c0.convertTo(kind=SpaceKind.BASIC)

    #     return c0

    def nearestOutcome(self, goal, context=None, n=1, bestContext=False, adaptContext=False, nearestUseContext=True, contextColumns=None):
        return self._nearestData(goal, context, n, bestContext, adaptContext, outcome=True, nearestUseContext=nearestUseContext,
            contextColumns=contextColumns)[:3]

    def _nearestData(self, goal, context=None, n=1, bestContext=False, adaptContext=False, outcome=True,
                     nearestUseContext=True, contextColumns=None, ignoreFirst=False, entity=None):
        self.outcomeSpace._validate()
        self.actionSpace._validate()
        if self.contextSpace:
            self.contextSpace._validate()
            self.outcomeContextSpace._validate()
            self.actionContextSpace._validate()

        if not outcome:
            contextColumns = self.contextColumns(contextColumns, goal, context)
        if not self.hasContext(self.contextSpace, contextColumns):
            context = None
        # if not outcome:
        #     print(context is None)
        #     print(contextColumns)
        if context:
            self.contextSpace._validate()
            # if adaptContext:
            #     context = self.adaptContext(goal, context=context)

        if outcome:
            goalSpace = self.outcomeSpace
            goalContextSpace = self.outcomeContextSpace
            otherSpace = self.actionSpace
        else:
            goalSpace = self.actionSpace
            goalContextSpace = self.actionContextSpace
            otherSpace = self.outcomeSpace

        goal = goal.projection(goalSpace)
        goalPlain = Data.npPlainData(goal, goalSpace)
        space = goalContextSpace if context else goalSpace
        context = context.convertTo(self.dataset, kind=SpaceKind.PRE).projection(
            self.contextSpace, entity=entity) if context else None
        goalContext = goal.extends(context)
        goalContextPlain = Data.npPlainData(goalContext, space)

        # Ids
        restrictionIds = self.restrictionIds
        restrictionIdsWithZeros = self.restrictionIds
        differentRestrictionIds = False

        # Remove zero values @TODO: discrete?
        # if outcome:
        #     restrictionIds = np.unique(np.nonzero(
        #         self.outcomeSpace.getData(restrictionIds))[0])
        
        # Remove zero values @TODO: discrete?
        minNorm = self.outcomeSpace.maxDistance * 0.001
        if outcome and goal.norm() > minNorm * 2:
            differentRestrictionIds = True
            restrictionIds = self.outcomeSpace.ids[np.argwhere(np.sum(np.abs(self.outcomeSpace.getData(restrictionIds)), axis=1) > minNorm)[:, 0]]

        if context:
            contextPlain = Data.npPlainData(context, self.contextSpace)
            numberContext = int(self.NN_CONTEXT + (len(restrictionIds) if restrictionIds is not None else self.contextSpace.number) * 0.04)
            # if bestContext:
            #     ids, _ = goalSpace.nearestPlain(goalPlain, n=numberContext,
            #                                restrictionIds=restrictionIds, otherSpace=otherSpace)
            #     cids, cdists = self.contextSpace.nearestDistancePlain(contextPlain, n=self.NN_LOCALITY,
            #                                                           restrictionIds=ids, otherSpace=otherSpace, columns=contextColumns)

            #     cdistMean = np.mean(cdists[:self.NN_LOCALITY // 2])

            #     if cdistMean >= self.contextSpace.maxDistance * 0.1:
            #         # print('Trying to find a better context')
            #         context = self.contextSpace.getPoint(cids)[0]
            #         goalContext = goal.extends(context)
            #         goalContextPlain = Data.npPlainData(goalContext, space)
            #         contextPlain = Data.npPlainData(context, self.contextSpace)
            # print(contextPlain)

            restrictionIds, _ = space.nearestDistancePlain(goalContextPlain,
                                                           n=numberContext,
                                                           restrictionIds=restrictionIds,
                                                           otherSpace=otherSpace,
                                                           columns=self.multiContextColumns(contextColumns, space, context),
                                                           adjustSpaceWeight=self.contextSpace,
                                                           adjustWeightFactor=10.)
            
            # restrictionIds, _ = self.contextSpace.nearestDistancePlain(contextPlain, n=numberContext,
            #                                                            restrictionIds=restrictionIds, otherSpace=otherSpace,
            #                                                            columns=contextColumns)

            if differentRestrictionIds:
                restrictionIdsWithZeros, _ = space.nearestDistancePlain(goalContextPlain,
                                                                        n=numberContext,
                                                                        restrictionIds=restrictionIdsWithZeros,
                                                                        otherSpace=otherSpace,
                                                                        columns=self.multiContextColumns(contextColumns, space, context),
                                                                        adjustSpaceWeight=self.contextSpace,
                                                                        adjustWeightFactor=10.)
            # print('===')
            # for id_ in restrictionIds:
            #     print(f'{self.actionSpace.getPoint(id_)[0]} + {self.contextSpace.getPoint(id_)[0]} -> {self.outcomeSpace.getPoint(id_)[0]}')
        
        # print(f'HEYHEY {goalContextPlain} {self.actionContextSpace.getNpPlainPoint(restrictionIds)}')

        # print('===')
        # print(D)
        # Compute best locality candidates
        # nearestUseContext = False
        useRestrictionIds = restrictionIds
        # useRestrictionIds = self.restrictionIds
        for i in range(2):
            if nearestUseContext:
                tids, tdist = space.nearestPlain(goalContextPlain,
                                            n=n+ignoreFirst,
                                            restrictionIds=useRestrictionIds,
                                            otherSpace=otherSpace,
                                            columns=self.multiContextColumns(contextColumns, space, context),
                                            adjustSpaceWeight=self.contextSpace)
            else:
                tids, tdist = goalSpace.nearestPlain(goalPlain,
                                                n=n+ignoreFirst,
                                                restrictionIds=useRestrictionIds,
                                                otherSpace=otherSpace)
            
            if ignoreFirst and len(tids) > 0:
                tids = tids[1:]
                tdist = tdist[1:]
            
            if i == 0:
                ids = tids
                dist = tdist
            else:
                ids = np.hstack([ids, tids])
                dist = np.hstack([dist, tdist])

            if context is None or useRestrictionIds is None:
                break
            # break
            # data = goalSpace.getData(ids)
            # distanceToGoal = np.mean(np.sum(np.abs(data - goalPlain), axis=1))
            break
            useRestrictionIds = self.restrictionIds
            # nearestUseContext = False
        
        # print('===')
        # # # print(distanceToGoal)
        # print(ids)
        # print(self.actionSpace.getData(ids))
        # print(self.contextSpace.getData(ids)[:, contextColumns])
        # print(self.outcomeSpace.getData(ids))
        # RegressionModel._db_ids = (ids, contextColumns)

        # print(ids)
        # print(dist)
        # print("Distances! 3 {}".format(
        #     [self.outcomeSpace.getPoint(id_)[0] for id_ in ids]))
        # print('---')
        # for id_ in ids:
        #     print(f'{id_} {self.actionSpace.getPoint(id_)[0]} + {self.contextSpace.getPoint(id_)[0]} -> {self.outcomeSpace.getPoint(id_)[0]}')
            # print(f'{self.actionSpace.getPoint(id_)[0]} -> {self.outcomeSpace.getPoint(id_)[0]}')

        number = len(ids)
        # print(self.actionSpace.getData(ids))
        # print(self.contextSpace.getData(ids)[:, 4])
        if outcome and number > 0:
            data = self.outcomeSpace.getData(ids)
            mean = np.mean(data, axis=0)
            for zero in (True, False):
                if zero:
                    distanceToCenter = np.sum(np.abs(data), axis=1)
                else:
                    distanceToCenter = np.sum(np.abs(data - mean), axis=1)
                indices = distanceToCenter > self.outcomeSpace.maxDistance * self.ALMOST_ZERO_FACTOR
                if np.sum(indices) < number * 3 // 10:
                    ids = ids[~indices]
                    dist = dist[~indices]
                    data = data[~indices]
                elif np.sum(~indices) < number * 3 // 10:
                    ids = ids[indices]
                    dist = dist[indices]
                    data = data[indices]
            # if np.sum(indices) > 0:
            #     data = data[indices]

        # Record ids
        self.lastCloseIds = ids

        if not differentRestrictionIds:
            restrictionIdsWithZeros = restrictionIds
        return ids, dist, context, restrictionIdsWithZeros, space, goal, goalPlain, goalContext, goalContextPlain, contextColumns

    def bestLocality(self, goal: Goal, context: Observation = None, getClosestOutcome=False, bestContext=False,
                     adaptContext=False, contextColumns=None, entity=None, dontMove=[]):
        """Compute most stable local action-outcome model around goal outcome."""
        results = self._nearestData(
            goal, context, self.NN_LOCALITY*2, bestContext, adaptContext, outcome=True, contextColumns=contextColumns, entity=entity)
        ids, dist, context, restrictionIds, space, goal, goalPlain, _, goalContextPlain, contextColumns = results

        # if context:
        #     adaptContext = False

        minPointsY = 5
        minPointsA = 5
        # minNorm = self.outcomeSpace.maxDistance * 0.001

        # print(f'=> {context} {self.outcomeSpace.getPlainPoint(ids)}')

        # Check if distance to goal is not too important
        if self.ENFORCE_MAX_NN_DISTANCE:
            idsPart = np.squeeze(np.argwhere(dist < self.outcomeSpace.maxNNDistance), axis=1)
            if len(idsPart) >= minPointsY:
                ids = ids[idsPart]

        # outcomes = self.outcomeSpace.getNpPlainPoint(ids)
        # idsPart = np.squeeze(np.argwhere(np.sum(outcomes, axis=1) > minNorm), axis=1)
        # if len(idsPart) < minPointsY:
        #     idsPart = range(0, len(idsPart))
        # ids = ids[idsPart]

        if len(ids) == 0:
            raise ActionNotFound("Not enough points to compute")

        if getClosestOutcome:
            distanceContext = (self.contextSpace.getPoint(restrictionIds[0])[0].distanceTo(context) /
                               self.contextSpace.maxDistance)
            y0 = self.outcomeSpace.getPoint(ids[0])[0]
            distanceGoal = goal.distanceTo(y0) / self.outcomeSpace.maxDistance
            # print(distanceContext)
            # print(distanceGoal)
            return y0, distanceGoal + distanceContext

        # actionSpace = self.actionContextSpace if adaptContext else self.actionSpace
        aList = self.actionSpace.getPlainPoint(ids)
        # print(f'==> {self.outcomeSpace.getPlainPoint(ids)}')

        bestScore = None
        scores = []
        results = []

        idsAs, distAs = self.actionSpace.nearestDistanceArray(aList, n=self.NN_ALOCALITY, otherSpace=space,
                                                              restrictionIds=restrictionIds)
        for i, (idsA, distA) in enumerate(zip(idsAs, distAs)):
            # print(len(idsA))
            # print(self.outcomeSpace.getNpPlainPoint(idsA))
            # print(distA)
            # idsA, distA = self.actionSpace.nearest(p, n=10, otherSpace=space, restrictionIds=self.restrictionIds)

            # Check if distance to a is too big and enough neighbours studied already
            # idsALarge = idsA[:]
            if self.ENFORCE_MAX_NN_DISTANCE:
                idsPart = np.squeeze(np.argwhere(distA < self.actionSpace.maxNNDistance), axis=1)
                if len(idsPart) >= minPointsA:
                    idsA = idsA[idsPart]
            if len(idsA) == 0:
                continue

            yPlain = self.outcomeSpace.getNpPlainPoint(idsA)
            aPlain = self.actionSpace.getNpPlainPoint(idsA)

            # if adaptContext:
            #     cPlain = self.contextSpace.getNpPlainPoint(idsA)

            if context:
                ycPlain = self.outcomeContextSpace.getNpPlainPoint(idsA)
                # acLargePlain = self.actionContextSpace.getNpPlainPoint(idsALarge)
            else:
                ycPlain = yPlain
                # acLargePlain = aLargePlain

            distanceGoal = 0.

            columns = self.multiContextColumns(
                contextColumns, self.outcomeContextSpace, context)
            a0Plain = multivariateRegression(
                ycPlain, aPlain, goalContextPlain, columns=columns)
            maxOutliers = self.actionSpace.maxDistance * self.OUTLIER_MAX_DISTANCE
            if np.any(a0Plain > maxOutliers):
                # print(f'/!!!!\\ {a0Plain} => {aPlain[-1]}\n{aPlain}')
                a0Plain = aPlain[-1]
            a0 = self.actionSpace.action(a0Plain).setRelative(True)


            # print(f'a{i}: {a0} |a|={a0.norm()} A YC GC\n{aPlain}\n{ycPlain}\n{goalContextPlain}')

            if a0.norm() > self.actionSpace.maxDistance:
                continue

            if self.actionSpace.primitive():
                a0 = a0.bounded()

            # actionCenter = np.mean(aPlain, axis=0)
            # actionCenterDistance = np.mean(
            #     np.sum((aPlain - actionCenter) ** 2, axis=1) ** .5)
            # actionDistance = np.mean(
            #     np.sum((aPlain - a0Plain) ** 2, axis=1) ** .5)
            # proximityScore = (
            #     actionDistance / (actionCenterDistance if actionCenterDistance != 0 else 1.) - 1.) / 20.

            # if context:
            #     a0 = self.actionSpace.action(a0Plain)
            #     ac0Plain = Data.plainData(a0.extends(context), self.actionContextSpace)
            # else:
            #     ac0Plain = a0Plain
            # y0Plain, error = multivariateRegressionError(acLargePlain, yLargePlain, ac0Plain)
            y0Plain, error = self.npForward(
                a0, context, contextColumns=contextColumns)
            goalDistance = euclidean(goalPlain, y0Plain)
            # print('---')
            # print(ycPlain)
            # print(aPlain)
            # print(goalContextPlain)
            # print('>')
            # print(a0Plain)
            # print(y0Plain)
            # print(goalDistance)

            # print(f'a{i}: {a0}; error: {error}; y0/goal: {y0Plain}=?{goalPlain}; goalDistance: {goalDistance}')

            score = goalDistance / self.outcomeSpace.maxDistance
                # 0.2 * proximityScore + 0.1 * error

            result = (a0, y0Plain, goalDistance, error, i)
            if dontMove:
                scores.append(score)
                results.append(result)
            elif bestScore is None or score < bestScore:
                bestScore = score
                results = [result]
        if not results:
            raise ActionNotFound(
                "Not enough points to compute action")
        
        if dontMove:
            results = np.array(results, dtype=object)
            # print(results)
            # print(np.argsort(scores))
            result = results[np.argsort(scores)]
            move = self.dataset.checkSpaceChanges(dontMove, results[0][0], context)
            if move >= 0.1:
                scores[0] += move
                for i, result in enumerate(results):
                    if i != 0:
                        scores[i] += self.dataset.checkSpaceChanges(dontMove, result[0], context)
                result = results[np.argsort(scores)]

        a0, y0Plain, goalDistance, error, i = results[0]
        self.bestLocalityCandidates[True if dontMove else False].append(i)
        # y0Plain, goalDistance, error = bestA0Plain, bestY0Plain, bestDistance, bestError

        y0 = self.outcomeSpace.asTemplate(y0Plain, entity=entity).setRelative(True)
        if entity:
            a0 = self.actionSpace.asTemplate(a0.plain(), entity=entity).setRelative(True)
        # if a0.length() > 1000:
        #     raise Exception()
        goalDistanceNormalized = goalDistance / self.outcomeSpace.maxDistance
        finalError = (error + goalDistanceNormalized) / 2
        return (a0, y0, context, self.computeCompetence(finalError),
                finalError, goalDistance, goalDistanceNormalized, error)



import math
import random
import numpy as np

import pygame
from pygame.locals import *
from pygame.color import *
import pymunk
from pymunk import Vec2d

from panda3d.core import Point2, Point3, Vec2, Vec3, Vec4
from dinos.environments.engines.tools.panda import PandaTools

from dinos.representation.physical_entity import PhysicalEntity
from dinos.representation.property import MethodObservable, AttributeObservable



class Wall(PhysicalEntity):
    def __init__(self, coordsFrom, coordsTo, width=40):
        super().__init__(self.__class__.__name__)
        self.width = width
        self.coordsFrom = coordsFrom
        self.coordsTo = coordsTo

        self.depth = 50
        self.pandaInit = False

    def initPhysics(self, physics):
        self.shape = pymunk.Segment(
            physics.static_body, self.coordsFrom, self.coordsTo, self.width // 2)
        self.shape.elasticity = 0.95
        self.shape.friction = 0.9
        physics.add(self.shape)

    def stopPhysics(self, physics):
        physics.remove(self.shape)

    def draw(self, base, drawOptions):
        line = self.shape
        if drawOptions.pygame:
            pygame.draw.line(base, THECOLORS["black"], (line.a.x, line.a.y), (line.b.x, line.b.y), self.width)
        if drawOptions.panda and not self.pandaInit:
            self.pandaInit = True
            a = Point2(line.a.x, line.a.y)
            b = Point2(line.b.x, line.b.y)
            print(a, b)
            v = b - a
            direction = direction = np.arctan2(v.y, v.x)# / np.math.pi * 180.

            w2 = self.width / 2
            pt = PandaTools()
            pt.color = Vec4(0.5, 0.5, 0.5, 1.)
            pt.quadri(Point3(a.x + np.cos(direction + np.math.pi / 2) * w2, a.y + np.sin(direction + np.math.pi / 2) * w2, self.depth),
                      Point3(a.x + np.cos(direction - np.math.pi / 2) * w2, a.y + np.sin(direction - np.math.pi / 2) * w2, self.depth),
                      Point3(b.x + np.cos(direction - np.math.pi / 2) * w2, b.y + np.sin(direction - np.math.pi / 2) * w2, self.depth),
                      Point3(b.x + np.cos(direction + np.math.pi / 2) * w2, b.y + np.sin(direction + np.math.pi / 2) * w2, self.depth), 0)
            node = pt.end()

            nodePath = base.render.attachNewNode(node)
            nodePath.setPos(0., 0., 0.)
            nodePath.setShaderAuto()



import math
import random
import numpy as np
from scipy.spatial.distance import euclidean

import pygame
from pygame.locals import *
from pygame.color import *
import pymunk
from pymunk import Vec2d
import pymunk.pygame_util

from direct.actor import Actor
from panda3d.core import Point2, Point3, Vec2, Vec3, Vec4, Material
from dinos.environments.engines.tools.panda import PandaTools

from dinos.representation.physical_entity import PhysicalEntity
from dinos.representation.property import MethodObservable, AttributeObservable, MethodEffector
from .cylinder import Cylinder


class Agent(Cylinder):
    def __init__(self, coords=(480, 300), radius=20, name='', omni=True, onlyX=False, xydiscretization=20):
        super().__init__(coords, radius, movable=True, name=name,
                         color=(0, 0, 0), xydiscretization=xydiscretization)
        self.markAsHost()

        self.onlyX = onlyX
        self.omni = omni

        self.mass = 2000.

        # MethodObservable(self, "position", self.absolutePosition)
        MethodObservable(self, "lidar", self.lidar,
                         discretization=50).learnable = False
        # AttributeObservable(self, "color", 'color')
        # AttributeObservable(self, "radius", 'radius')
        # AttributeObservable(self, "mass", 'mass')
        actions = {
            'down': (0., 1.0),
            'up': (0., -1.0),
            'left': (-1.0, 0.),
            'right': (1.0, 0.),
        }
        MethodEffector(self, "move", self.move,
                       dim=1 if self.onlyX else 2, actions=actions)
        # MethodEffector(self, "toggle", self.toggle, dim=0)

        self.sufaceWidth = radius * 2 + 10
        self.surface = pygame.Surface(
            (self.sufaceWidth, self.sufaceWidth), pygame.SRCALPHA)

    def lidar(self, property=None):
        return self.lidarFromPos(self.shape, self.engine.physics, self.shape.body.position, self.onlyX)

    @staticmethod
    def lidarFromPos(ownShape, physics, position, onlyX):
        # Distances[0 E, 1 SE, 2 S, 3 SW, 4 W, 5 NW, 6 N, 7 NE]
        number = 2 if onlyX else 8
        maxdist = 70
        width = 10

        dirs = np.linspace(0, 2*np.pi, number + 1)[:-1]
        distances = np.zeros(number)
        for i, dir in enumerate(dirs):
            d = Vec2d(maxdist, 0.0).rotated(dir)
            hits = physics.segment_query(position, position + d,
                                                width, pymunk.ShapeFilter())
            dists = [info.alpha *
                     maxdist for info in hits if info.shape != ownShape]
            distances[i] = min(dists) if dists else maxdist
        return distances

    def toggle(self, parameters, property=None):
        self.shape.body.position = (
            self.shape.body.position.x + 20, self.shape.body.position.y)

    def move(self, parameters, property=None):
        # print(f'MOVE! {parameters}')
        self.performAction(self._move, args=(
            parameters, property), duration=0.6, step=0.05)

    def _move(self, parameters, property=None):
        # print("MOVE")
        # print(self.absolutePosition())
        # print(parameters)
        wl = min(max(parameters[0], -1), 1)  # ** 3
        wr = 0. if self.onlyX else min(max(parameters[1], -1), 1)  # ** 3
        noise = 0.
        pw = 10000

        # self.body.apply_impulse_at_local_point(Vec2d(1.0, 0.0) * pw * float(wl), (0, 0))

        if self.omni:
            self.body.apply_impulse_at_local_point(Vec2d(1.0, 0.0) * pw * float(wl + random.uniform(-noise, noise)) +
                                                   Vec2d(0.0, 1.0) * pw * float(wr + random.uniform(-noise, noise)),
                                                   (0, 0))
            if wl != 0. or wr != 0.:
                self.direction = np.arctan2(wl, wr)
        else:
            pass

    def draw(self, base, drawOptions):
        if drawOptions.pygame:
            pygame.draw.circle(base, (224, 224, 0), list(
                map(int, self.absolutePosition())), self.radius, 1)

            pygame.draw.circle(self.surface, (224, 224, 224),
                            (self.sufaceWidth // 2, self.sufaceWidth // 2), self.radius)
            pygame.draw.circle(self.surface, self.color, (self.sufaceWidth //
                                                        2, self.sufaceWidth // 2), self.radius, 1)

            # Orientation
            w = 10
            points = [(self.sufaceWidth // 2 + self.radius, self.sufaceWidth // 2),
                    (self.sufaceWidth // 2 + self.radius -
                    w, self.sufaceWidth // 2 - w // 2),
                    (self.sufaceWidth // 2 + self.radius - w, self.sufaceWidth // 2 + w // 2)]
            pygame.draw.polygon(self.surface, self.color, points)

            # Wheels
            for pos in (-1, 1):
                pygame.draw.line(self.surface, self.color,
                                (self.sufaceWidth // 2 - self.radius // 2,
                                self.sufaceWidth // 2 + pos * self.radius // 2),
                                (self.sufaceWidth // 2 + self.radius // 2, self.sufaceWidth // 2 + pos * self.radius // 2), 3)

            surface = pygame.transform.rotate(
                self.surface, np.rad2deg(self.direction) - 90.)
            base.blit(surface, list(
                map(lambda x: int(x) - surface.get_width() // 2, self.absolutePosition())))
        if drawOptions.panda:
            if not self.pandaInit:
                self.pandaInit = True

                self.robotNode = base.render.attachNewNode('robot axis')
                # self.robotModel = Actor.Actor('panda-model', {'walk': 'panda-walk4'})
                self.robotModel = base.loader.loadModel('rb1.bam')
                self.robotModel.reparentTo(self.robotNode)
                self.robotModel.setScale(self.radius / 0.8)
                # self.robotModel.setScale(0.5)
                # self.pandaWalk = self.robotModel.actorInterval('walk', playRate=1.8)
                # self.pandaWalk.loop()

                print(self.robotModel.findAllMaterials())

                material = Material()
                material.setShininess(5.0)
                material.setBaseColor((0, 0, 1, 1))
                # material.setAmbient((0, 0, 1, 1))
                self.robotModel.setMaterial(material)
            self.robotNode.setPos(*self.body.position, 10.)
            self.robotNode.setHpr(-np.rad2deg(self.direction) - 90, 0., 0.)




import copy

from exlab.utils.io import parameter
from dinos.data.data import Data


class State(object):
    def __init__(self, environment, values=[], dataset=None, featureMap=None, enforceDataset=True):
        self.environment = environment
        self.dataset = dataset
        self.featureMap = featureMap
        if not isinstance(values, list):
            values = [values]
        self.values = [part for value in values for part in value.flat()]
        if dataset and enforceDataset:
            self.values = [part.convertTo(dataset) for part in self.values]
        self.update()

    def __repr__(self):
        return f"State ({self.values})"

    def __iter__(self):
        return self.values.__iter__()

    def update(self):
        if self.featureMap:
            self.featureMap.update(self.values, self.dataset)
        self._context = Data(*self.values)
    
    def consequences(self, controllable, excludeSpaces=None):
        excludeSpaces = list(parameter(excludeSpaces, set()))

        consequences = []
        controllables = [controllable]
        while controllables:
            newControllables = []
            for controllable in controllables:
                newControllables += self._controllableConsequences(controllable, excludeSpaces)
            controllables = newControllables
            consequences +=  controllables
        
        return consequences
        

    def _controllableConsequences(self, controllable, excludeSpaces):
        consequences = []
        controllableSpace = controllable.space
        context = self.context()

        for model in self.dataset.enabledModels():
            if model.isCoveredByActionSpaces(controllableSpace):
                result, _ = model.forward(controllable, context)
                if result:
                    for part in result.flat():
                        if part.space not in excludeSpaces:
                            consequences.append(part)
                            excludeSpaces.append(part)

        return consequences

    def apply(self, controllable, excludeSpaces=set(), overwrite=None):
        consequences = self.consequences(controllable, excludeSpaces)

        variations = {c.space: c for c in consequences}

        for i, part in enumerate(self.values):
            if part.space in variations.keys():
                self.values[i] = part + variations[part.space]
            if overwrite:
                overpart = overwrite.projection(part.space)
                if overpart:
                    self.values[i] = overpart



        # variations = {}
        # actionSpace = controllable.space
        # context = self.context()
        # for model in dataset.enabledModels():
        #     if model.isCoveredByActionSpaces(actionSpace):
        #         result, _ = model.forward(controllable, context)
        #         if result:
        #             parts = result.flat()
        #             for part in parts:
        #                 variations[part.space] = part
        #         # else:
        #         #     # print(f'======= {model} {action} {context}')
        #         #     result[0].flat()

        # for i, part in enumerate(self.values):
        #     if part.space in variations.keys():
        #         self.values[i] = part + variations[part.space]
        #     if overwrite:
        #         overpart = overwrite.projection(part.space)
        #         if overpart:
        #             self.values[i] = overpart

        self.update()
        return self

    def context(self):
        return self._context

    def difference(self, previous, space=None):
        diff = self._context.difference(previous._context)
        if space:
            diff = diff.projection(diff)
        return diff

    def copy(self):
        return self.__class__(self.environment, list(self.values), self.dataset, self.featureMap, enforceDataset=False)



import math
import random
import numpy as np
from scipy.spatial.distance import euclidean

import pygame
from pygame.locals import *
from pygame.color import *
import pymunk
from pymunk import Vec2d
import pymunk.pygame_util

from panda3d.core import Point2, Point3, Vec2, Vec3, Vec4
from dinos.environments.engines.tools.panda import PandaTools

from dinos.representation.physical_entity import PhysicalEntity
from dinos.representation.property import MethodObservable, AttributeObservable


class Cylinder(PhysicalEntity):
    def __init__(self, coords=(480, 300), radius=20, movable=True, name='', color=None, xydiscretization=20):
        super().__init__(self.__class__.__name__, name)
        self.coordsInit = coords
        self.coords = coords
        self.direction = 0.

        self.pandaInit = False

        self.radius = radius
        self.color = color if color else (
            (0, 224, 128) if movable else (255, 0, 0))
        self.mass = 1. if movable else 1000000.

        MethodObservable(self, 'position', self.absolutePosition,
                         discretization=xydiscretization)
        MethodObservable(self, 'positionToAgent', self.positionToAgent,
                         discretization=xydiscretization)  # .learnable = False
        
        # Visual properties have to be the same for all entities!
        AttributeObservable(self, 'color', 'color', discretization=128, visual=True)
        # AttributeObservable(self, 'radius', 'radius', discretization=10)
        # AttributeObservable(self, 'mass', 'mass', discretization=100)

    # def relativePosition(self, property=None):
    #     # - self.env.agents[0].shape.body.position
    #     relativePosition = self.body.position
    #     return [relativePosition.x, relativePosition.y]

    def absolutePosition(self, property=None):
        return [self.shape.body.position.x, self.shape.body.position.y]

    def positionToAgent(self, property=None):
        relativePosition = self.body.position - \
            self.world.child('Agent').shape.body.position
        return [relativePosition.x, relativePosition.y]

    # def move(self, parameters, property=None):
    #     # print("MOVE")
    #     # print(parameters)
    #     wl = min(max(parameters[0], -1), 1) ** 3
    #     wr = min(max(parameters[1], -1), 1) ** 3
    #     noise = 0.
    #
    #     self.body.apply_impulse_at_local_point(Vec2d(1.0, 0.0) * 1500 * float(wl + random.uniform(-noise, noise)) +
    #                                            Vec2d(0.0, 1.0) * 1500 * float(wr + random.uniform(-noise, noise)),
    #                                            (0, 0))

    def initPhysics(self, physics):
        inertia = pymunk.moment_for_circle(self.mass, 0, self.radius, (0, 0))
        body = pymunk.Body(self.mass, inertia)
        body.position = self.coords[0], self.coords[1]
        shape = pymunk.Circle(body, self.radius, (0, 0))
        shape.elasticity = 0.95
        physics.add(body, shape)
        self.shape = shape
        self.body = body

    def stopPhysics(self, physics):
        physics.remove(self.body, self.shape)

    def draw(self, base, drawOptions):
        if drawOptions.pygame:
            pygame.draw.circle(base, self.color, list(map(int, self.absolutePosition())), self.radius)
        if drawOptions.panda:
            if not self.pandaInit:
                self.pandaInit = True

                pt = PandaTools()
                pt.color = Vec4(*self.color, 1.)
                pt.cylinder(self.radius, 40.)
                node = pt.end()

                self.nodePath = base.render.attachNewNode(node)
                self.nodePath.setShaderAuto()
            self.nodePath.setPos(*self.body.position, 0.)
            



'''
    File name: dataset.py
    Author: Alexandre Manoury
    Python Version: 3.6
'''

import numpy as np

from exlab.modular.module import manage

from dinos.data.spacemanager import SpaceManager


class Dataset(SpaceManager):
    """The dataset used to record actions, outcomes and procedures."""

    def __init__(self, options={}):
        """
        options dict: parameters for the dataset
        """
        super().__init__('dataset', storesData=True, options=options)
        self.logger.tag = 'dataset'
        self.learner = None

    def _serialize(self, serializer):
        dict_ = super()._serialize(serializer)
        dict_.update(serializer.serialize(
            self, ['options']))
        return dict_

    @classmethod
    def _deserialize(cls, dict_, serializer, obj=None):
        if obj is None:
            obj = cls(dict_.get('options', {}))
        return super()._deserialize(dict_, serializer, obj)

    def _postDeserialize(self, dict_, serializer):
        super()._postDeserialize(dict_, serializer)
        serializer.set('dataset', self)
        serializer.set('dataset', self, category='spaceManager')
    
    def attachLearner(self, learner):
        self.learner = learner
        manage(self).attach(learner)



import sys
import copy
import math
import random
import numpy as np
from scipy.spatial import cKDTree
from scipy.spatial.distance import euclidean

# import matplotlib as plt

# from ..utils.logging import DataEventHistory, DataEventKind
# from ..utils.io import getVisual, plotData, visualize, concat
# from ..utils.serializer import Serializable, Serializer
# from ..utils.maths import multivariateRegression, multivariateRegressionError, multivariateRegression_vector, popn
from exlab.utils.io import parameter
from exlab.interface.serializer import Serializable


from dinos.data.data import *
from dinos.data.space import SpaceKind
# from ..data.abstract import *
from dinos.data.path import ActionNotFound
from dinos.data.contextarea import ContextSpatialization


'''

'''


class Model(Serializable):
    number = 0
    ALMOST_ZERO_FACTOR = 0.002
    PRECISION_LEARNING_RATE = 0.01
    PRECISION_MULTIPLIER = 2.
    PRECISION_GOAL_NUMBER = 25

    def __init__(self, dataset, actionSpace, outcomeSpace, contextSpace=[], restrictionIds=None, model=None,
                 register=True, metaData={}):
        self.id = Model.number
        Model.number += 1

        self.dataset = dataset
        self.enabled = True

        self.allowCompetenceCaching = True
        self.invalidateCompetences()

        self.createdSince = -1
        self.lowCompetenceSince = -1
        self.evaluations = {}
        self.attemptedContextSpaces = {}
        self.precision = [-1., -1.]
        self.precisionPerGoalNorm = np.zeros(self.PRECISION_GOAL_NUMBER)
        self.precisionPerGoalNormNumber = np.zeros(self.PRECISION_GOAL_NUMBER)
        self.limitMoves = 1.

        self.restrictionIds = restrictionIds
        # self.nonDuplicateLastId = -1
        # self.nonDuplicateActionIds = []
        # self.nonDuplicateOutcomeIds = []

        self.actionSpace = dataset.multiColSpace(actionSpace)
        self.outcomeSpace = dataset.multiColSpace(outcomeSpace)
        self.contextSpace = dataset.multiColSpace(contextSpace)
        self.contextSpace = self.contextSpace.convertTo(self.dataset, kind=SpaceKind.PRE)

        self.actionContextSpace = dataset.multiColSpace(
            [self.actionSpace, self.contextSpace], weight=0.5)
        self.outcomeContextSpace = dataset.multiColSpace(
            [self.outcomeSpace, self.contextSpace], weight=0.5)

        self.contextSpacialization = None
        if self.contextSpace:
            self.contextSpacialization = [ContextSpatialization(self, self.actionSpace)] #, ContextSpatialization(self, self.contextSpace)]

        # self.spacesHistory = DataEventHistory()

        if register:
            self.dataset.registerModel(self)

    def __repr__(self):
        disabled = '❌' if not self.enabled else ''
        return f'{disabled}Model{self.__class__.__name__}({self.actionSpace} | {self.contextSpace} => {self.outcomeSpace})'
    
    def _sid(self, serializer):
        return serializer.serialize(self, foreigns=['dataset', 'actionSpace', 'outcomeSpace', 'contextSpace'], reference=True)

    def _serialize(self, serializer):
        dict_ = serializer.serialize(
            self, ['enabled', 'createdSince', 'lowCompetenceSince', 'contextSpacialization', 'precision'],
            foreigns=['dataset', 'actionSpace', 'outcomeSpace', 'contextSpace'])
        return dict_

    @classmethod
    def _deserialize(cls, dict_, serializer, obj=None):
        if obj is None:
            obj = cls._findOrCreate(serializer.deserialize(dict_.get('dataset')),
                    serializer.deserialize(dict_.get('actionSpace')),
                    serializer.deserialize(dict_.get('outcomeSpace')),
                    serializer.deserialize(dict_.get('contextSpace')),
                    dict_.get('metaData', {}))
        return super()._deserialize(dict_, serializer, obj)

    @classmethod
    def _findOrCreate(cls, dataset, actionSpace, outcomeSpace, contextSpace=[], metaData={}):
        obj = None
        if obj is None:
            obj = cls(dataset, actionSpace, outcomeSpace, contextSpace, metaData=metaData)
        return obj

    def _postDeserialize(self, dict_, serializer):
        super()._postDeserialize(dict_, serializer)
        for attr in ['enabled', 'createdSince', 'lowCompetenceSince', 'precision']:
            setattr(self, attr, dict_.get(attr))
        for i, csd in enumerate(dict_.get('contextSpacialization', [])):
            if i < len(self.contextSpacialization):
                serializer.deserialize(csd, obj=self.contextSpacialization[i])

    # @classmethod
    # def _deserialize(cls, dict_, dataset, spaces, loadResults=True, options={}, obj=None):
    #     a = [next(i for i in spaces if i.name == name or i.id == name)
    #          for name in dict_['actions']]
    #     y = [next(i for i in spaces if i.name == name or i.id == name)
    #          for name in dict_['outcomes']]
    #     c = [next(i for i in spaces if i.name == name or i.id == name)
    #          for name in dict_.get('context', [])]
    #     obj = cls(dataset, a, y, c)
    #     return obj

    # def actualOutcomeContextSpace(self, contextColumns):
    #     if not np.any(contextColumns):
    #         return self.outcomeSpace
    #     return self.outcomeContextSpace
    
    # def actualActionContextSpace(self, contextColumns):
    #     if not np.any(contextColumns):
    #         return self.actionSpace
    #     return self.actionContextSpace

    @classmethod
    def adaptiveEpisode(cls, adaptiveModelManager, events):
        pass

    def justCreated(self):
        pass

    @property
    def duration(self):
        if self.createdSince == -1:
            return 0
        return self.dataset.learner.iteration - self.createdSince
    
    def matches(self, model, ignoreContext=False):
        return (self.actionSpace == model.actionSpace and
                self.outcomeSpace == model.outcomeSpace and
                (ignoreContext or self.contextSpace == model.contextSpace or (not self.contextSpace and not model.contextSpace)))
    
    def includes(self, model, ignoreContext=False):
        return (self.actionSpace.includes(model.actionSpace) and
                self.outcomeSpace.includes(model.outcomeSpace) and
                (ignoreContext or (not self.contextSpace and not model.contextSpace) or self.contextSpace.includes(model.contextSpace)))

    def update(self):
        pass

    def metaData(self):
        return None

    def continueFrom(self, previousModel):
        self.id = previousModel.id
        previousModel.id = -1

        self.createdSince = previousModel.createdSince
        self.lowCompetenceSince = previousModel.lowCompetenceSince
        self.evaluations = previousModel.evaluations

        # if self.contextSpacialization and previousModel.contextSpacialization:
        #     self.contextSpacialization[0].continueFrom(
        #         previousModel.contextSpacialization[0])

        # previousModel.spacesHistory.extend(self.spacesHistory)
        # self.spacesHistory = previousModel.spacesHistory
    
    def findActionFromEvent(self, event):
        if self.isCoveredByActionSpaces(event.actions.space):
            return event.actions.projection(self.actionSpace)
        elif self.isCoveredByActionSpaces(event.primitiveActions.space):
            return event.primitiveActions.projection(self.actionSpace)
        else:
            return event.outcomes.projection(self.actionSpace)

    def pointAdded(self, event):
        if self.contextSpacialization:
            # print(event.allActions)
            # outcome = event.outcomes.projection(self.outcomeSpace)
            self.contextSpacialization[0].update(event.allActions)

            # context = event.context.projection(self.contextSpace)
            # self.contextSpacialization[1].addPoint(context, event.iteration)

    def updatePrecision(self, success, distanceToGoal, weight=1, index=0):
        # oldp = self.precision[index]
        precision = min(distanceToGoal * self.PRECISION_MULTIPLIER, self.outcomeSpace.maxDistance * 0.5)
        if self.precision[index] < 0:
            self.precision[index] = self.outcomeSpace.maxDistance * 0.03
        learningRate = self.PRECISION_LEARNING_RATE * weight
        if distanceToGoal < self.precision[index]:
            if not success:
                return
            learningRate *= 2
        self.precision[index] = self.precision[index] * (1 - learningRate) + (precision - self.precision[index]) * learningRate
        if index == 1 and self.precision[index] < self.precision[0]:
            self.precision[index] = self.precision[0]
        self.precision[index] = min(max(self.precision[index], self.outcomeSpace.maxDistance * 0.005), self.outcomeSpace.maxDistance * 0.5)
        # print(f'#{index} from {oldp} to {self.precision[index]} // {distanceToGoal} // {self}')
    
    def getPrecision(self, ratio=0.01, multiplier=1):
        if self.precision[0] < 0:
            return self.outcomeSpace.maxDistance * ratio
        return self.precision[0] * multiplier
    
    def updatePrecisionPerGoal(self, goalMove, error):
        index = int(goalMove.norm() / self.outcomeSpace.maxDistance * self.PRECISION_GOAL_NUMBER)
        precision = min(error, self.outcomeSpace.maxDistance * 0.5)
        if index >= self.PRECISION_GOAL_NUMBER:
            return
        # learningRate = self.PRECISION_LEARNING_RATE * 5
        self.precisionPerGoalNormNumber[index] = min(self.precisionPerGoalNormNumber[index] + 1, 100)
        self.precisionPerGoalNorm[index] = (
            (self.precisionPerGoalNorm[index] * (self.precisionPerGoalNormNumber[index] - 1)) + precision) / self.precisionPerGoalNormNumber[index]
        
        # print(f'UPDATING {self} {goalMove.norm()} => {index}: {error}')
    
    def explorable(self):
        return self.precision[0] >= 0.
        #  and self.precision[0] <= self.outcomeSpace.maxDistance * 0.25
    
    def maxDistance(self, outcome):
        return (self.outcomeSpace.maxDistance / 4 if self.outcomeSpace.maxDistance != 0 else 1.) + outcome.norm() / 2

    # Context
    def hasContext(self, contextSpace, contextColumns):
        return contextSpace and (contextColumns is None or np.any(contextColumns))

    def contextColumns(self, contextColumns, action, context):
        if contextColumns is not None:
            return contextColumns
        if not self.contextSpacialization:
            return None
        return self.contextSpacialization[0].columns(action)# & self.contextSpacialization[1].columns(context)

    def multiContextColumns(self, contextColumns, space, context):
        if contextColumns is None or not np.any(contextColumns) or context is None:
            return None
        indices = space.columnsFor(self.contextSpace)
        cols = np.full(space.dim, True)
        cols[indices] = contextColumns
        return cols
    
    def retrieveContext(self, context, entity=None):
        return context.convertTo(self.dataset, kind=SpaceKind.PRE).projection(self.contextSpace, entity=entity)

    # Operations
    def forward(self, action: Action, context: Observation = None, contextColumns=None, ignoreFirst=False, entity=None):
        value, error = self.npForward(action, context, contextColumns=contextColumns, ignoreFirst=ignoreFirst, entity=entity)
        if value is None or np.isnan(np.sum(value)):
            return None, 1
        return Data(self.outcomeSpace.applyTo(entity), value.tolist()).setRelative(True), error

    def npForward(self, action: Action, context: Observation = None, contextColumns=None, ignoreFirst=False, entity=None):
        raise NotImplementedError()

    def inverse(self, goal: Goal, context: Observation = None, contextColumns=None, entity=None):
        raise NotImplementedError()

    def bestLocality(self, goal: Goal, context: Observation = None, contextColumns=None, entity=None, dontMove=[]):
        raise NotImplementedError()

    def computeCompetence(self, error):
        error = min(error, 2)
        return max(0, min(1., (1. - error ** 2) / np.exp((error * 20))))  # ** 3
    
    # def computeCompetence(self, error, distanceGoal=0):
    #     distanceGoal = min(distanceGoal, 1.)
    #     return max(0, min(1, (1. - distanceGoal - error) / np.exp((error * 4.15) ** 3)))

    def getIds(self):
        return self.findSharedIds(self.outcomeSpace, self.actionSpace, self.contextSpace, restrictionIds=self.restrictionIds)
    
    # def updateNonDuplicateIds(self):
    #     number = self.actionSpace.number
    #     if self.nonDuplicateLastActionId < number:
    #         self.actionSpace.findDuplicates(self.nonDuplicateActionIds)

    #         self.nonDuplicateLastActionId < number

    #     self.nonDuplicateLastId = id_
    #     self.nonDuplicateActionIds = []
    #     self.nonDuplicateOutcomeIds = []
    
    @staticmethod
    def findSharedIds(self, *spaces, restrictionIds):
        assert len(spaces) > 0, 'At least 1 space is required!'

        spaces = list(spaces)
        ids = spaces.pop(0).getIds()
        if restrictionIds:
            ids = np.intersect1d(ids, restrictionIds)
        for space in spaces:
            ids = np.intersect1d(ids, space.getIds())
        return ids

    def competence(self, precise=False, onlyIds=None, contextColumns=None, contextColumnsOverwrite=None):
        if onlyIds is None and contextColumns is None:
            if self._lastCompetencePrecise:
                return self._lastCompetencePrecise
            elif not precise and self._lastCompetence:
                return self._lastCompetence

        c = self.computeCompetence(self.std(precise=precise, onlyIds=onlyIds, contextColumns=contextColumns, contextColumnsOverwrite=contextColumnsOverwrite))

        if onlyIds is None and contextColumns is None:
            if precise:
                self._lastCompetencePrecise = c
            else:
                self._lastCompetence = c
        return c

    def competenceData(self, actions, outcomes, contexts=None, precise=False, contextColumns=None):
        return self.computeCompetence(self.stdData(actions, outcomes, contexts, precise=precise, contextColumns=contextColumns))

    def invalidateCompetences(self):
        self._lastCompetence = None
        self._lastCompetencePrecise = None
    
    def performant(self, competence, duration):
        return self.competence(precise=True) >= competence and self.duration >= duration

    # data=None, context: Observation = None,
    def std(self, precise=False, onlyIds=None, contextColumns=None, contextColumnsOverwrite=None):
        errorsZeros, _, _ = self._errorForwardAll(
            precise=precise, almostZeros=True, onlyIds=onlyIds, contextColumns=contextColumns, contextColumnsOverwrite=contextColumnsOverwrite)
        errorZeros = np.mean(errorsZeros) if len(errorsZeros) > 0 else 0.

        errorsNonZeros, _, _ = self._errorForwardAll(
            precise=precise, almostZeros=False, onlyIds=onlyIds, contextColumns=contextColumns, contextColumnsOverwrite=contextColumnsOverwrite)
        errorNonZeros = np.mean(errorsNonZeros) if len(errorsNonZeros) > 0 else 0.

        #np.percentile(errorsZeros, 65)
        # print('---', errorZeros, errorNonZeros)

        return 1 - (1 - errorNonZeros) * (1 - errorZeros)
    
    def stdData(self, actions, outcomes, contexts=None, precise=False, onlyIds=None, contextColumns=None):
        errors, _ = self._errorForwardDataAll(
            actions, outcomes, contexts, precise=precise, contextColumns=contextColumns)
        if len(errors) == 0:
            return 0.
        return np.mean(errors)
    
    # def variance(self, data=None, context: Observation = None, precise=False, contextColumns=None):
    #     return self.std(data=data, context=context, precise=precise, contextColumns=contextColumns) ** 2

    def forwardErrors(self, precise=False, almostZeros=None, onlyIds=None, contextColumns=None, contextColumnsOverwrite=None, sortIds=False):
        errors, _, ids = self._errorForwardAll(precise=precise, almostZeros=almostZeros, onlyIds=onlyIds,
                                            contextColumns=contextColumns, contextColumnsOverwrite=contextColumnsOverwrite)
        if sortIds:
            order = np.argsort(-errors)
            distanceMax = (self.outcomeSpace.maxDistance / 4 if self.outcomeSpace.maxDistance != 0 else 1.)
            trueErrors = errors * distanceMax
            return errors[order], ids[order], trueErrors[order]
        return errors

    def forwardError(self, eventId, contextColumns=None, contextColumnsOverwrite=None):
        action = self.actionSpace.getPoint(eventId)[0]
        outcome = self.outcomeSpace.getPoint(eventId)[0]
        context = self.contextSpace.getPoint(eventId)[0] if self.contextSpace else None

        return self._errorForwardData(action, outcome, context, contextColumns=contextColumns, contextColumnsOverwrite=contextColumnsOverwrite, eventId=eventId)
    
    def _errorForwardData(self, action, outcome, context=None, contextColumns=None, contextColumnsOverwrite=None, eventId=None):
        if not self.contextSpace:
            context = None
        if contextColumns is not None and contextColumnsOverwrite is not None:
            contextColumns[contextColumnsOverwrite] = self.contextSpacialization[0].columns(action)[contextColumnsOverwrite]
        # print(f'{self.actionSpace.getPoint(eventId)[0]} + {self.contextSpace.getPoint(eventId)[0]} -> {self.outcomeSpace.getPoint(eventId)[0]}')

        # contextColumns = self.contextColumns(contextColumns, action, context)

        # context = context if self.hasContext(
        #     self.contextSpace, contextColumns) else None

        # if context is None:
        #     contextColumns = None

        # print(contextColumns)
        outcomeEstimated, linearError = self.forward(action, context, contextColumns=contextColumns, ignoreFirst=False)
        # acPlain = action.extends(context).npPlain()
        # outcomeEstimated = self.outcomeSpace.point(self.fnn.predict([acPlain])[0])
        if not outcomeEstimated:
            return 1.

        # print(outcome)
        # print(outcomeEstimated)
        # print(linearError)
        errorOutcome = outcomeEstimated.distanceTo(outcome) / self.maxDistance(outcome) + min(linearError, 1.) * 0.1

        errorOutcome = min(errorOutcome, 1.)
        # print(f'{action}, {outcome} ?= {outcomeEstimated} -> {errorOutcome}')
        # self.dataset.logger.info(f'{errorOutcome:.2f} #{eventId}: {action.plain()} + {context} -> {outcome.plain()} vs estimated {outcomeEstimated.plain()}')

        # if errorOutcome > 0.1 or linearError > 0.1:
        #     if context:
        #         context = context.plain()
        #     print(
        #         f'Failed {errorOutcome:.3f} {linearError:.3f} #{eventId}: {action.plain()} + {context} -> {outcome.plain()} vs estimated {outcomeEstimated.plain()}')
            # print('--- ERROR ---')
            # print(outcome)
            # print(outcomeEstimated)
            # print(errorOutcome)
            # print(context)
            # print(action)
            # print('---       ---')
        return errorOutcome, linearError

    def _errorForwardAll(self, precise=False, almostZeros=None, onlyIds=None, contextColumns=None, contextColumnsOverwrite=None):
        ids = parameter(onlyIds, self.getIds())
        if almostZeros is not None:
            data = self.outcomeSpace.getData(ids)
            indices = np.sum(np.abs(data), axis=1) > self.outcomeSpace.maxDistance * self.ALMOST_ZERO_FACTOR
            if almostZeros:
                indices = ~indices
            if np.sum(indices) == 0:
                return [], [], []
            ids = ids[indices]
        if not precise:
            number = min(80 if not almostZeros else 50, self.outcomeSpace.number // 10)
            ids_ = np.arange(len(ids))
            np.random.shuffle(ids_)
            ids = ids[ids_[:number]]
        errors = np.array([self.forwardError(id_, contextColumns=contextColumns, contextColumnsOverwrite=contextColumnsOverwrite) for id_ in ids]).T
        return errors[0], errors[1], ids

    def _errorForwardDataAll(self, actions, outcomes, contexts=None, precise=False, exceptAlmostZero=False, contextColumns=None):
        # data = np.hstack((actions, outcomes, contexts) if contexts is not None else (actions, outcomes))
        ids = np.arange(len(actions))
        if exceptAlmostZero:
            indices = np.sum(
                np.abs(outcomes), axis=1) > self.outcomeSpace.maxDistance * self.ALMOST_ZERO_FACTOR
            if np.sum(indices) == 0:
                return []
            ids = ids[indices]

        if not precise:
            number = min(100, self.outcomeSpace.number // 10)
            ids_ = np.arange(len(ids))
            np.random.shuffle(ids_)
            ids = ids[ids_[:number]]

        if contexts is not None:
            data = zip(actions, outcomes, contexts)
        else:
            data = zip(actions, outcomes)
        errors = np.array([self._errorForwardData(*d, contextColumns=contextColumns) for d in data]).T
        return errors[0], errors[1]

    # Space coverage
    def reachesSpaces(self, spaces):
        spaces = self.dataset.convertSpaces(spaces)
        return self.outcomeSpace.intersects(spaces)

    def __spaceCoverage(self, spaces, selfSpaces, covers=0):
        selfSpaces = set(selfSpaces.flatSpaces)
        if not isinstance(spaces, list):
            spaces = [spaces]
        spaces = self.dataset.convertSpaces(spaces)
        spaces = [space for space in spaces if space]
        spaces = set([sp for space in spaces for sp in space.flatSpaces])
        intersects = selfSpaces.intersection(spaces)
        if covers == 0:
            return intersects == spaces
        if covers == 1:
            return len(intersects) > 0
        else:
            return intersects == selfSpaces

    def coversActionSpaces(self, spaces):
        return self.__spaceCoverage(spaces, self.actionSpace, 0)

    def intersectsActionSpaces(self, spaces):
        return self.__spaceCoverage(spaces, self.actionSpace, 1)

    def isCoveredByActionSpaces(self, spaces):
        return self.__spaceCoverage(spaces, self.actionSpace, 2)

    def coversOutcomeSpaces(self, spaces):
        return self.__spaceCoverage(spaces, self.outcomeSpace, 0)

    def intersectsOutcomeSpaces(self, spaces):
        return self.__spaceCoverage(spaces, self.outcomeSpace, 1)

    def isCoveredByOutcomeSpaces(self, spaces):
        return self.__spaceCoverage(spaces, self.outcomeSpace, 2)

    def coversContextSpaces(self, spaces):
        return self.__spaceCoverage(spaces, self.contextSpace, 0)

    def intersectsContextSpaces(self, spaces):
        return self.__spaceCoverage(spaces, self.contextSpace, 1)

    def isCoveredByContextSpaces(self, spaces):
        return self.__spaceCoverage(spaces, self.contextSpace, 2)

    def currentContext(self, env):
        return env.observe(spaces=self.contextSpace).convertTo(self.dataset)

    def controllableContext(self):
        return self.dataset.controllableSpaces(self.contextSpace, merge=True)

    def nonControllableContext(self):
        return self.dataset.nonControllableSpaces(self.contextSpace, merge=True)
    
    # Visual

    '''def _draw_graph(self):
        """
        Draw Graph
        """
        plt.clf()
        if rnd is not None:
            plt.plot(rnd[0], rnd[1], "^k")
        for node in self.nodeList:
            if node.parent is not None:
                plt.plot([node.x, self.nodeList[node.parent].x], [
                         node.y, self.nodeList[node.parent].y], "-g")

        for (ox, oy, size) in self.obstacleList:
            plt.plot(ox, oy, "ok", ms=30 * size)

        plt.plot(self.start.x, self.start.y, "xr")
        plt.plot(self.end.x, self.end.y, "xr")
        plt.axis([-2, 15, -2, 15])
        plt.grid(True)'''

    # Visual
    # def get_action_outcomes_visualizer(self, prefix="", context=False):
    #     """Return a dictionary used to visualize outcomes reached for the specified outcome space."""
    #     def onclick(event):
    #         center = (event['event'].xdata, event['event'].ydata)
    #         y0 = self.estimate_goal(center)[0]
    #         competence, std, dist = self.get_competence_std(center)
    #         print("{} -> {} {}".format(center, y0, competence))

    #     '''ids = []
    #     datao = self.outcomeSpace.getData()
    #     for d, id_ in zip(datao, self.outcomeSpace.ids):
    #         if d[0] ** 2 + d[1] ** 2 < 0.1:
    #             ids.append(id_)
    #     idsc = list(ids)
    #     for id_ in idsc:
    #         d = self.actionSpace.getPlainPoint([id_])
    #         if d.tolist():
    #             d = d[0]
    #             if d[0] ** 2 + d[1] ** 2 < 0.1:
    #                 ids.remove(id_)
    #         else:
    #             ids.remove(id_)
    #     data = self.actionSpace.getPlainPoint(ids)
    #     print(ids)
    #     return getVisual(
    #                     [lambda fig, ax, options: plotData(data, fig, ax, options)],
    #                     minimum=self.actionSpace.bounds['min'],
    #                     maximum=self.actionSpace.bounds['max'],
    #                     title=prefix + "Points competence for " + str(self)
    #                     )'''

    #     spaces = [self.actionSpace, self.outcomeSpace]
    #     if context:
    #         spaces.append(self.contextSpace)
    #     plots = [space.getPointsVisualizer(prefix) for space in spaces]
    #     plots[1]['onclick'] = onclick
    #     return plots

    # def get_competence_visualizer(self, dataset=None, prefix=""):
    #     data = np.array(dataset) if dataset else self.outcomeSpace.getData()
    #     print("Len {}".format(len(data)))
    #     competences = np.array([self.get_competence_std(d)[0] for d in data])
    #     ids = np.squeeze(np.argwhere(competences >= 0))
    #     competences = competences[ids]
    #     data = data[ids]
    #     print("Mean {}".format(np.mean(competences)))

    #     def onclick(event):
    #         center = (event['event'].xdata, event['event'].ydata)
    #         y0 = self.estimate_goal(center)[0]
    #         competence, std, dist = self.get_competence_std(center)
    #         print("{} -> {} {}".format(center, y0, competence))
    #     #competences = (competences - np.min(competences)) / (np.max(competences) - np.min(competences))
    #     return getVisual(
    #         [lambda fig, ax, options: plotData(
    #             data, fig, ax, options, color=competences)],
    #         minimum=self.outcomeSpace.bounds['min'],
    #         maximum=self.outcomeSpace.bounds['max'],
    #         title=prefix + "Points competence for " + str(self),
    #         colorbar=True,
    #         onclick=onclick
    #     )

    # # Plot
    # def plot(self):
    #     visualize(self.get_action_outcomes_visualizer())

    # # Api
    # def apiget(self, range_=(-1, -1)):
    #     return {'spaces': self.spacesHistory.get_range(range_), 'interest': self.interestMap.apiget(range_)}



import numpy as np
import random
import copy
import math
import time
import threading

from exlab.lab.counter import AsyncCounter
from exlab.modular.module import Module, manage
from exlab.interface.serializer import Serializer

from exlab.utils.io import parameter
from dinos.utils.move import MoveConfig
from dinos.utils.database import Database

from dinos.agents.tools.planners.planner import Planner
from dinos.agents.tools.performers.performer import Performer
from dinos.agents.tools.strategies.strategy_set import StrategySet


# def make(id_, environment):
#     from ..utils.loaders import DataManager
#     return DataManager.makeAgent(id_, environment)


class Agent(Module):
    """An Agent performing actions in an environment.
    Can be subclassed to implement an algorithm designed to perform one or multiple tasks.

    Args:
        environment (Environment): The environment within which the agent will live and operate
        performer (Performer):
        options (dict): A dictionary of parameters for the agent

    Attributes:
        dataset (Dataset): Dataset used by the agent, when learning data
        reachStrategies (Strategy[]):
        iteration (int):
        lastIterationTime (int):
        iterationTimes (float[]):
        environment (Environment)
        options (dict)
        performer (Performer)
    """

    DISCRETE_STATES = False
    DISCRETE_ACTIONS = False

    PLANNER_CLASS = Planner
    PERFORMER_CLASS = Performer

    def __init__(self, host, dataset=None, performer=None, planner=None, options={}):
        super().__init__('Agent', host.manager, loggerTag='agent')
        manage(self).attach_counter(AsyncCounter(self))

        self.host = host
        self.host.hosting = self

        self.environment = host.world.manager
        self.assertDiscrete(self.environment)

        self.dataset = dataset
        self.featureMap = None
        self.options = options
        self.scheduled = None
        self.iterationEvent = threading.Event()

        self.learningPolicy = None

        self.configs = {}
        self.analyse = self.__class__.ResultAnalyser(self)

        # self.iteration = 0
        self.episode = 0
        self.iterationByEpisode = []
        self.iterationType = {}

        self.lastIterationEpisode = 0

        self.testStrategies = StrategySet(agent=self)

        self.performer = parameter(performer, self.PERFORMER_CLASS(self, options=options))
        self.planner = parameter(planner, self.PLANNER_CLASS(self, options=options))

        # debug metrics
        self.lastIterationTime = -1
        self.iterationTimes = []

    def __repr__(self):
        return f'Agent {self.__class__.__name__}'

    def _serialize(self, serializer):
        """Returns a dictionary with serialized information.

        Args:
            options (dict): Parameters of the serialization

        Returns:
            type: Serialized Object
        """
        dict_ = super()._serialize(serializer)
        dict_.update(serializer.serialize(
            self, ['options', 'testStrategies', 'iteration', 'configs'], foreigns=['host']))
        return dict_
    
    @classmethod
    def _deserialize(cls, dict_, serializer, obj=None):
        if obj is None:
            obj = cls(serializer.find(dict_.get('host')),
                      serializer.deserialize(dict_.get('dataset')),
                      options=dict_.get('options', {}))
        
        return super()._deserialize(dict_, serializer, obj)
    
    def _postDeserialize(self, dict_, serializer):
        super()._postDeserialize(dict_, serializer)
        env = serializer.get('environment')
        if env.iteration < dict_.get('iteration', 0):
            env.counter.iteration = dict_.get('iteration')
        self.syncCounter()

        self.configs = serializer.deserialize(dict_.get('configs'), default={})
    
    def deserializer(self):
        d = self.environment.deserializer()
        d.set('agent', self)
        if self.dataset:
            d.set('dataset', self.dataset)
            d.set('dataset', self.dataset, category='spaceManager')
        d.attach_finder('strategy', self.findStrategy)
        return d

    def findStrategy(self, name):
        return None

    # @classmethod
    # def _deserialize(cls, dict_, environment, options={}, obj=None):
    #     """Deserialized a dict into an object.

    #     Args:
    #         dict_ (dict): Object data
    #         environment (Environment): The environment within which the agent will live and operate
    #         options (dict): Deserializing parameters
    #         obj (Agent): When subclassing, the object created by the deserialize method of the subclass
    #     """
    #     from ..utils.loaders import DataManager

    #     obj = obj if obj else cls(
    #         environment, options=dict_.get('options', {}))
    #     obj = Module._deserialize(dict_, options=options, obj=obj)

    #     for strategy in dict_.get('testStrategies', []):
    #         obj.testStrategies.add(DataManager.loadType(strategy['path'], strategy['type'])
    #                                .deserialize(strategy, obj, options=options))
    #     return obj

    def currentConfig(self, name=''):
        d = {
            'environment': self.environment.__class__.__name__,
            'learner': self.__class__.__name__,
            'name': name,
        }
        return d

    def save(self, name):
        serializer = Serializer()
        data = {'agent': self.serialize(serializer)}
        evaluator = self.environment.evaluator(self)
        if evaluator:
            data['evaluator'] = evaluator.serialize(serializer)
        db = Database.from_data(self.currentConfig(name), data)
        db.save()
        return db
    
    def load(self, path):
        db = Database.from_file(Database.databasedir / path)
        db.load()
        d = self.deserializer()
        if self.dataset:
            d.deserialize(db.data.get('agent', {}).get('dataset', {}), obj=self.dataset)
        self._postDeserialize(db.data.get('agent', {}), d)

        d.deserialize(db.data.get('evaluator', {}), obj=self.environment.evaluator(self, create=True))
    
    @classmethod
    def loadAgent(self, environment, path):
        db = Database.from_file(Database.databasedir / path / Database.FILENAME)
        db.load()
        d = environment.deserializer()
        agent = d.deserialize(db.data.get('agent', {}))
        d.deserialize(db.data.get('evaluator', {}), obj=environment.evaluator(agent, create=True))
        return agent

    @property
    def iteration(self):
        return manage(self).counter.t

    @property
    def counter(self):
        return manage(self).counter
    
    def syncCounter(self):
        self.counter.sync()

    def trainable(self):
        return False

    def testable(self):
        return True

    @property
    def discreteStates(self):
        return self.DISCRETE_STATES

    @property
    def discreteActions(self):
        return self.DISCRETE_ACTIONS

    def assertDiscrete(self, environment, autochange=True):
        if self.discreteStates and not environment.discreteStates or self.discreteActions and not environment.discreteActions:
            if not environment.CAN_BE_DISCRETIZED:
                autochange = False
            if not autochange:
                raise Exception('Trying to apply a discrete learner to a continuous environment!\n' +
                                'Check if you have forgotten to call env.discretizeStates/Actions = True?')
            if self.discreteStates and not environment.discreteStates:
                environment.discretizeStates = True
            if self.discreteActions and not environment.discreteActions:
                environment.discretizeActions = True
    
    def actions(self):
        return self.host.actions()

    def explorableSpaces(self, onlyPrimitives=False):
        spaces = [a.space for a in self.host.actions()]
        if not onlyPrimitives and self.dataset:
            spaces += self.dataset.controllableSpaces(explorable=True)
        return list(set(spaces))
    
    def observe(self, formatParameters=None):
        if self.featureMap:
            self.featureMap.populateFrom(self.environment)
        return self.host.observeFrom(formatParameters=formatParameters)
    
    def schedule(self, method, *args, **kwargs):
        if self.environment.threading:
            def func():
                method(*args, **kwargs)
                self.scheduled = None

            self.scheduled = func
        else:
            method(*args, **kwargs)

    def reach(self, configOrGoal=MoveConfig()):
        if not isinstance(configOrGoal, MoveConfig):
            configOrGoal = MoveConfig(goal=configOrGoal)
        self.test(configOrGoal)

    def test(self, config=MoveConfig()):
        if config.goal and self.dataset:
            config.goal = self.dataset.convertData(config.goal)
        config.exploitation = True

        self.logger.debug2(f'Testing {config}')

        self.syncCounter()
        self.iterationType[self.iteration] = 'evaluation' if config.evaluating else 'test'
        self.schedule(self._test, config)

    def _test(self, config):
        self.syncCounter()
        strategy = self.testStrategies.sample()
        config.strategy = strategy
        memory = strategy.run(config)
        self.syncCounter()
        self.iterationType[self.iteration] = 'end'

        self._postTest(memory, config)
    
    def _postTest(self, memory, config):
        pass

    def perform(self, action):
        self.performer.performActions(action)

    def _performAction(self, action, config=MoveConfig()):
        self.host.scheduledAction = True
        self.environment.execute(action, config=config, agent=self, sync=True)

    def step(self, action, countIteration=True):
        result = self.environment.step(action)
        if countIteration:
            self.iteration += 1
        return result
    
    def propertySpace(self, filter_=None, kind=None):
        return self.environment.propertySpace(filter_, kind=kind, dataset=self.dataset)
    
    def getIterationType(self, iteration):
        last = None
        for i in self.iterationType:
            if i > iteration:
                if last is None or self.iterationType[last] == 'end':
                    return ''
                return self.iterationType[last]
            last = i
        return ''

    class ResultAnalyser(object):
        def __init__(self, agent):
            self.agent = agent

    # def addReachStrategy(self, strategy):
    #     """Add a Strategy designed to perform a task in a certain way.
    #
    #     Args:
    #         strategy (Strategy): The given Strategy, should possesses a reach method
    #     """
    #     if strategy in self.reachStrategies:
    #         return
    #     assert strategy.agent == self
    #     self.reachStrategies.append(strategy)
    #     self.addChildModule(strategy)
    #
    # def reach(self, goalOrConfig, config=MoveConfig()):
    #     """Tries to reach a given goal with constraints.
    #
    #     Args:
    #         goalOrConfig (Goal | MoveConfig): The Goal to reach (or a MoveConfig describing the Goal to reach)
    #         config (MoveConfig): Configuration for executing the task
    #     """
    #     if isinstance(goalOrConfig, MoveConfig):
    #         config = goalOrConfig
    #     else:
    #         config.goal = goalOrConfig
    #     if not self.reachStrategies:
    #         raise Exception('{} has no strategy for reaching a goal'.format(self))
    #     if self.dataset:
    #         config.goal = config.goal.convertTo(self.dataset)
    #     config.exploitation = True
    #     return self.reachStrategies[0].reach(config)
    #
    # def measureIterationTime(self):
    #     """Measures the execution time at each iteration
    #     """
    #     currentTime = int(round(time.time() * 1000))
    #     if self.lastIterationTime > 0:
    #         self.iterationTimes.append((self.iteration, currentTime - self.lastIterationTime))
    #     self.lastIterationTime = currentTime



import math
import random
import numpy as np
from scipy.spatial.distance import euclidean

import pygame
from pygame.locals import *
from pygame.color import *
import pymunk
from pymunk import Vec2d
import pymunk.pygame_util

from dinos.representation.live_entity import LiveEntity
from dinos.representation.property import MethodObservable, AttributeObservable


class Button(LiveEntity):
    def __init__(self, coords, radius=20, name='', xydiscretization=20):
        super().__init__(self.__class__.__name__, name)
        self.coordsInit = coords
        self.coords = Vec2d(coords)

        self.radius = radius

        MethodObservable(self, 'pressed', self.pressed,
                         discretization=1)

        # Visual properties have to be the same for all entities!
        # AttributeObservable(self, 'pressed', 'pressed',
        #                     discretization=1, visual=True)
        # AttributeObservable(self, 'radius', 'radius', discretization=10)
        # AttributeObservable(self, 'mass', 'mass', discretization=100)

    # def relativePosition(self, property=None):
    #     # - self.env.agents[0].shape.body.position
    #     relativePosition = self.body.position
    #     return [relativePosition.x, relativePosition.y]

    def pressed(self, property=None):
        for cylinder in self.world.cascadingChildren('Cylinder'):
            if self.coords.get_distance(Vec2d(cylinder.absolutePosition())) < cylinder.radius + self.radius:
                return True
        return False
        # return [self.shape.body.position.x, self.shape.body.position.y]

    def draw(self, base, drawOptions):
        if drawOptions.pygame:
            pygame.draw.circle(base, (0, 0, 200), (int(self.coords.x), int(self.coords.y)), self.radius)



import sys
import copy
import math
import random
import numpy as np
from scipy.spatial.distance import euclidean

from exlab.interface.serializer import Serializable


class Data(Serializable):
    def __init__(self, *args):
        if len(args) == 0:
            # Null data
            self.space = None
            self.value = []
            self._valueOrdered = []
            self._parts = []
        else:
            if len(args) == 2 and not isinstance(args[0], Data):
                # Data(space, value)
                self.space = args[0]
                value = args[1].tolist() if isinstance(args[1], np.ndarray) else list(args[1])
                if len(self.space.cols) == 1:
                    self._parts = [SingleData(self.space, value)]
                else:
                    self._parts = []
                    for part in self.space:
                        self._parts.append(SingleData(part, value[:part.dim]))
                        del value[:part.dim]
            else:
                # Data(Data#0, Data#1, ...)
                self._parts = list(args)
                self.space = self._parts[0].space.spaceManager.multiColSpace(
                    [p.space for p in self._parts])

            self.abstract = any(p.abstract for p in self._parts)
            self.value = self.plain()
            self._valueOrdered = None

    def _serialize(self, serializer):
        dict_ = serializer.serialize(self, ['_parts'])
        return dict_

    @classmethod
    def _deserialize(cls, dict_, serializer, obj=None):
        if obj is None:
            obj = cls(*[serializer.deserialize(part) for part in dict_.get('parts', [])])

        return super()._deserialize(dict_, serializer, obj)

    # @classmethod
    # def _deserialize(cls, dict_, options={}, obj=None):
    #     obj = obj if obj else cls(dict_.get('iteration'), dict_.get('meanError'), dict_.get('meanStd'),
    #                               dict_.get('results'), dict_.get('models'))
    #     return obj

    # Operators
    def __iter__(self):
        return self._parts.__iter__()

    def __getitem__(self, arg):
        return self._parts[arg]

    def __len__(self):
        return len(self._parts)

    def __bool__(self):
        return len(self._parts) > 0

    def __neg__(self):
        return self.__class__(*(-i for i in self))

    def __mul__(self, value):
        return self.__class__(*(v*value for v in self))

    def __add__(self, other):
        return self.__class__(*(v + other.singleSpaceComposante(v.space) for v in self))

    def __sub__(self, other):
        return self.__class__(*(v - other.singleSpaceComposante(v.space) for v in self))

    def __eq__(self, other):
        if isinstance(other, Data):
            if self.space is None or other.space is None:
                return self.space == other.space
            return (self.space == other.space or self.space.native == other.space.native) and self.valueOrdered == other.valueOrdered
        return False
    
    def distanceTo(self, other):
        if not isinstance(other, Data) or not self.space.matches(other.space, kindSensitive=False):
            raise ValueError('Spaces should be the same for both data')

        return euclidean(self.valueOrdered, other.valueOrdered)

    def difference(self, previous):
        result = []
        for point in self.flat():
            if point.space.relativeLearning:
                result.append(
                    point - previous.singleSpaceComposante(point.space))
            else:
                result.append(point)
        return Observation(*result)

    def norm(self):
        return np.linalg.norm(self.value)
    
    def norm1(self):
        return np.sum(np.abs(self.value))

    def length(self):
        return np.linalg.norm(self.value)
    
    @property
    def valueOrdered(self):
        if self._valueOrdered is None:
            self._valueOrdered = self.plainOrdered()
        return self._valueOrdered

    def setRelative(self, relative=None):
        if relative is not None:
            for part in self._parts:
                part.setRelative(relative)
        return self

    def absoluteData(self, state):
        return self.__class__(*[part.absoluteData(state) for part in self._parts])

    def relativeData(self, state, ignoreRelative=False):
        return self.__class__(*[part.relativeData(state, ignoreRelative=ignoreRelative) for part in self._parts])
    
    def clone(self):
        return self.__class__(*[part.clone() for part in self._parts])

    # Flatting
    def flat(self):
        return [x for p in self._parts for x in p.flat()]

    def plain(self):
        return [x for p in self._parts for x in p.plain()]

    def plainOrdered(self, spaceOrder=None, assertSize=True, fill=False):
        if not self._parts:
            return []
        if len(self._parts) == 1:
            return self.value

        if not spaceOrder:
            spaceOrder = self.space

        # parts = [x for s in spaceOrder.flatColsWithMultiRows for x in self.singleSpaceComposante(s).plainOrdered()]
        if self.abstract:
            parts = [x for s in spaceOrder.flatColsWithMultiRows for x in self.singleSpaceComposante(
                s).plainOrdered()]
        else:
            parts_ = [[list(map(self.singleSpaceComposante, possibility))
                       for possibility in group] for group in spaceOrder.groupedCols]
            parts = []
            for group in parts_:
                if len(group) == 1:
                    parts += group[0]
                else:
                    for possibility in group:
                        if len([p for p in possibility if p != Data()]) == len(possibility):
                            parts += possibility
                            break
            parts = [x for part in parts for x in part.plainOrdered()]

        if assertSize and len(parts) != spaceOrder.dim:
            raise Exception(f"Size mismatch while plaining out data {self} into {spaceOrder}: {len(parts)} != {spaceOrder.dim}")
        return parts

    def npPlain(self, spaceOrder=None, assertSize=True, fill=False):
        return np.array(self.plain())

    # Projections
    def singleSpaceComposante(self, singleSpace, kindSensitive=False, entity=None):
        flatten = self.flat()
        # print(flatten)
        pm = next((part for part in flatten if part.space.findMatchingSpaceRows(
            singleSpace, idSensitive=True, entity=entity)), Data())
        if not pm:
            pm = next((part for part in flatten if part.space.findMatchingSpaceRows(
                singleSpace, entity=entity)), Data())
        if not pm and not kindSensitive:
            pm = next((part for part in flatten if part.space.findMatchingSpaceRows(
                singleSpace, kindSensitive=False, entity=entity)), Data())
        pm = pm.applyTo(entity)
        return pm

    def projection(self, space, allowAbstraction=True, kindSensitive=False, entity=None):
        parts = [self.singleSpaceComposante(s, kindSensitive=kindSensitive, entity=entity)
                 for s in space.flatColsWithMultiRows]
        return self.__class__._vectorClass_(*[p for p in parts if p != Data()])

    # Entity
    # def _findSpaceEntity(self, space, entity):
    #     if not space.rowAggregation:
    #         return self.singleSpaceComposante(space)
    #     parts = [[p for p in possibility if p.boundProperty.entity == entity]
    #              for possibility in space.groupedCols[0]]
    #     part = [
    #         possibility for group in parts for possibility in group if possibility][0]
    #     return self.__class__._vectorClass_(part, np.array(self.valueOrdered)[self.space.columnsFor(space)])

    def applyTo(self, entity):
        # TODO: can only be applied to one entity abstraction, needs to be extended
        # if not self.abstract:
        #     return self
        parts = [part.applyTo(entity) for part in self]
        return self.__class__._vectorClass_(*parts)

    # Changing data
    def extends(self, other):
        if other is None:
            return self
        assert isinstance(other, Data)
        return self.__class__(*(self._parts + other._parts))

    def update(self, data, kindSensitive=True):
        if len(self) == 0:
            return self

        flatten = self.flat()
        flattenNew = data.convertTo(self._parts[0].space.spaceManager).flat()

        for newPart in flattenNew:
            for i, part in enumerate(flatten):
                if part.space.matches(newPart.space, kindSensitive=kindSensitive):
                    flatten[i] = newPart.__class__(
                        newPart.space, newPart.value)
                    break

        return self.__class__._vectorClass_(*[p for p in flatten])
    
    def convertTo(self, spaceManager=None, kind=None, toData=None):
        return self.__class__(*[part.convertTo(spaceManager=spaceManager, kind=kind, toData=toData)
                                for part in self._parts])


    def asTemplate(self, values):
        return self.__class__(self.space, values)
    
    def bounded(self, bounds=(-1., 1.)):
        return self.__class__(*[part.bounded(bounds) for part in self._parts])

    @staticmethod
    def plainData(obj, spaceOrder=None, assertSize=True, fill=False):
        if isinstance(obj, SingleData):
            return list(obj.value)
        if isinstance(obj, Data):
            return obj.plainOrdered(spaceOrder, assertSize=assertSize, fill=fill)
        if isinstance(obj, list) and obj and (isinstance(obj[0], Data) or isinstance(obj[0], SingleData)):
            return [Data.plainData(o, spaceOrder=spaceOrder, assertSize=assertSize, fill=fill) for o in obj]
        return obj

    @staticmethod
    def npPlainData(obj, spaceOrder=None, assertSize=True, fill=False):
        val = Data.plainData(obj, spaceOrder=spaceOrder,
                             assertSize=assertSize, fill=fill)
        if isinstance(val, list):
            return np.array(val)
        return val

    def toStr(self, short=False):
        if short:
            return '\n    '.join([p.toStr() for p in self._parts])
        return f"{self.__class__.__name__}[{self.toStr(short=True)}]"

    def __repr__(self):
        return self.toStr()


class SingleData(Data):
    """
    Represents a space data
    :param space: data space
    :param value: data value
    """

    def __init__(self, space, value):
        self.space = space
        self.value = value if isinstance(value, list) or isinstance(value, np.ndarray) else [value]
        self._valueOrdered = self.value
        self._parts = [self]
        if len(self.value) != self.space.dim:
            raise Exception(f"Size mismatch: space {self.space} is {self.space.dim}d and the data value {self.value} is {len(self.value)}d")
        self.relative = self.space.relative
        self.abstract = self.space.abstract
    
    def _serialize(self, serializer):
        dict_ = serializer.serialize(self, ['value', 'relative'], foreigns=['space'])
        return dict_

    @classmethod
    def _deserialize(cls, dict_, serializer, obj=None):
        if obj is None:
            obj = cls(serializer.deserialize(dict_.get('space')), dict_.get('value'))

        return super()._deserialize(dict_, serializer, obj)
    
    def _postDeserialize(self, dict_, serializer):
        super()._postDeserialize(dict_, serializer)
        if dict_.get('relative'):
            self.setRelative(dict_.get('relative'))

    """
    Get the associated space (no CompositeSpace)
    :return SingleSpace instance
    """
    # def get_space(self):
    #     return self.space

    """
    Check whether this data is a vector or not
    :return False
    """

    def isVector(self):
        return False

    # Operators
    def __neg__(self):
        return self.__class__(self.space, [-v for v in self.value]).setRelative(self.relative)

    def __mul__(self, value):
        return self.__class__(self.space, [v*value for v in self.value]).setRelative(self.relative)

    def __add__(self, other):
        if other.space is None:
            raise ValueError('Cannot add data with none data')
        if not self.space.matches(other.space, kindSensitive=False, rowMatching=True):
            raise ValueError(f'Spaces (or native spaces) should be the same for both data (first is in {self.space.name} and second is in {other.space.name})')
        d = self.__class__(
            self.space, [v1 + v2 for v1, v2 in zip(self.value, other.value)])
        d.setRelative(self.relative and other.relative)
        return d

    def __sub__(self, other):
        if other.space is None:
            raise ValueError('Cannot add data with none data')
        if not self.space.matches(other.space, kindSensitive=False, rowMatching=True):
            raise ValueError(f'Spaces (or native spaces) should be the same for both data (first is in {self.space.name} and second is in {other.space.name})')
        d = self.__class__(
            self.space, [v1 - v2 for v1, v2 in zip(self.value, other.value)])
        d.setRelative(self.relative and other.relative)
        return d
    
    def setRelative(self, relative=None):
        if relative is not None:
            self.relative = relative
        return self
    
    def absoluteData(self, state):
        value = self.value
        if self.relative:
            for s in state:
                if s.space.matches(self.space):
                    value = [v1 + v2 for v1, v2 in zip(self.value, s.value)]
                    break
        return self.__class__(self.space, value).setRelative(False)

    def relativeData(self, state, ignoreRelative=False):
        value = self.value
        if not self.relative or ignoreRelative:
            for s in state:
                if s.space.matches(self.space):
                    value = [v1 - v2 for v1, v2 in zip(self.value, s.value)]
                    break
        return self.__class__(self.space, value).setRelative(True)
    
    def clone(self):
        return self.__class__(self.space, self.value).setRelative(self.relative)

    # Flatting
    def flat(self):
        return [self]

    def plain(self):
        return self.value

    def plainOrdered(self):
        return self.value

    # Changing data
    def convertTo(self, spaceManager=None, kind=None, toData=None):
        spaceManager = spaceManager if spaceManager else self.space.spaceManager
        return self.__class__(spaceManager.convertSpace(self.space, kind=kind, toData=toData),
                              self.value).setRelative(self.relative)
    
    def applyTo(self, entity):
        if not entity:
            return self
        row = self.space.applyTo(entity)
        if not row:
            return self
        return self.__class__(row, self.value).setRelative(self.relative)
    
    def bounded(self, bounds=(-1., 1.)):
        value = [min(max(v, bounds[0]), bounds[1]) for v in self.value]
        return self.__class__(self.space, value).setRelative(self.relative)

    def toStr(self, short=False):
        ls = ', '.join(["{: .3f}".format(v) for v in self.value])
        return f"{self.space.toStr(True)}{'±' if self.relative else ''}{ls}"

    def __repr__(self):
        return f"{self.__class__.__name__}({self.toStr(True)})"


class DataSequence(Serializable):
    """
    Represents a sequence data
    :param parts, ...: the sub data items
    """

    def __init__(self, *args):
        self._parts = list(args)

    def _serialize(self, serializer):
        dict_ = serializer.serialize(self, ['_parts'])
        return dict_

    # Operators
    def append(self, part):
        self._parts.append(part)

    def __iter__(self):
        return self._parts.__iter__()

    def __getitem__(self, arg):
        return self._parts[arg]

    def __len__(self):
        return len(self._parts)
    
    def clone(self):
        return self.__class__(*[part.clone() for part in self._parts])

    def __repr__(self):
        return f'{self.__class__.__name__}{self._parts}'

    def convertTo(self, spaceManager=None, kind=None, toData=None):
        return self.__class__(*[part.convertTo(spaceManager, kind, toData) for part in self._parts])


class Goal(Data):
    """
    Represents a Goal
    """
    pass


class SingleAction(SingleData):
    """
    Represents an Action in 1 SingleSpace
    """
    pass


# Action : set of simulatneous sub actions
class Action(Data):
    """
    Represents an Action in multiple Spaces
    """
    pass
    #def __repr__(self):
    #return "{}({})".format(self.__class__.__name__, '; '.join(["{}:{}".format(a.space.id, a.value) for a in self.flat()]))


# ActionList : set of successives actions
class ActionList(DataSequence):
    """
    Represents a sequence of ActionList in multiple Spaces
    """
    pass


class SingleObservation(SingleData):
    """
    Represents an SingleObservation in 1 SingleSpace
    """
    pass


class Observation(Data):
    """
    Represents an SingleObservation in multiple Spaces
    """
    pass


class Point(Data):
    """
    Represents an SingleObservation in multiple Spaces
    """
    pass


Data._vectorClass_ = Data
Observation._vectorClass_ = Observation
SingleObservation._vectorClass_ = Observation
Action._vectorClass_ = Action
SingleAction._vectorClass_ = Action




from .agent import Agent
from dinos.agents.tools.strategies.random import RandomStrategy


class RandomAgent(Agent):
    def __init__(self, host, dataset=None, performer=None, planner=None, options={}):
        super().__init__(host, dataset=dataset, performer=performer,
                         planner=planner, options=options)
        self.testStrategies.add(RandomStrategy(self))



from .pymunk import PyMunkFeatureMap
from dinos.environments.playground import Agent


class PlaygroundFeatureMap(PyMunkFeatureMap):
    def __init__(self):
        super().__init__()
        self.addPropertyGenerator('Agent', 'lidar', self.lidar)

    def lidar(self, entity, property=None):
        return Agent.lidarFromPos(entity.shape, self.physics, entity.shape.body.position, entity.native.onlyX)



# import numpy as np
# import random
# import pygame
# from pygame.locals import *
# from pygame.color import *

from dinos.environments.engines.internals.pymunk import PymunkEngine
from dinos.environments.environment import Environment

# from worlds.pandaworld import PandaWorld
# from ..scene import SceneSetup, Challenge
# from ..testbench import *
# from ...utils.objects import MoveConfig

# from .objects import *


class PlaygroundEnvironment(Environment):
    NAME = 'Playground'
    DESCRIPTION = 'A simple room with a wheeled robot and some obstacles'
    ENGINE = PymunkEngine

    # def __init__(self, scene=None, options={}):
    #     super().__init__(scene, options)
